# modal tester, use when necessary
from textual.app import App, ComposeResult
from textual.widgets import Button

from rovr.screens import FileInUse


class Test(App):
    CSS_PATH = "../style.tcss"

    def compose(self) -> ComposeResult:
        yield Button("hi")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.push_screen(
            FileInUse(
                "The file appears to be open in another application and cannot be deleted.\nPath: file.txt",
            ),
            lambda x: self.notify(str(x)),
        )


Test().run()



from os import getcwd, path

from textual import work
from textual.widgets import Button

from rovr.classes import (
    EndsWithAnArchiveExtension,
    EndsWithRar,
    IsValidFilePath,
    PathDoesntExist,
)
from rovr.functions.icons import get_icon
from rovr.functions.path import normalise
from rovr.screens import ModalInput
from rovr.variables.constants import config


class ZipButton(Button):
    ALLOW_MAXIMIZE = False

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            get_icon("general", "zip")[0], classes="option", id="zip", *args, **kwargs
        )

    def on_mount(self) -> None:
        if config["interface"]["tooltips"]:
            self.tooltip = "Compress selected files"

    @work
    async def on_button_pressed(self, event: Button.Pressed) -> None:
        if self.disabled:
            return
        selected_files = await self.app.query_one("#file_list").get_selected_objects()
        if not selected_files:
            self.notify(
                "No files selected to zip.",
                title="Zip Files",
                severity="warning",
            )
            return

        parent_folder_name = path.basename(getcwd())
        default_zip_name = f"{parent_folder_name}.zip"

        response: str = await self.app.push_screen(
            ModalInput(
                border_title="Create Zip Archive",
                border_subtitle="Enter the name for the zip file",
                initial_value=default_zip_name,
                validators=[
                    PathDoesntExist(strict=False),
                    IsValidFilePath(),
                    EndsWithRar(),
                    EndsWithAnArchiveExtension(),
                ],
                is_path=True,
            ),
            wait_for_dismiss=True,
        )

        if not response:
            return

        archive_name = normalise(path.join(getcwd(), response))

        self.app.query_one("ProcessContainer").create_archive(
            selected_files, archive_name
        )
        self.app.query_one("#file_list").focus()



from textual import events
from textual.app import ComposeResult
from textual.containers import Container, Grid
from textual.screen import ModalScreen
from textual.widgets import Button, Label

from rovr.variables.constants import config


class DeleteFiles(ModalScreen):
    """Screen with a dialog to confirm whether to delete files."""

    def __init__(self, message: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.message = message

    def compose(self) -> ComposeResult:
        with Grid(id="dialog"):
            yield Label(self.message, id="question")
            if config["settings"]["use_recycle_bin"]:
                yield Button("\\[D] Trash", variant="warning", id="trash")
                yield Button("\\[X] Delete", variant="error", id="delete")
                with Container():
                    yield Button("\\[C]ancel", variant="primary", id="cancel")
            else:
                yield Button("\\[X] Delete", variant="error", id="delete")
                yield Button("\\[C]ancel", variant="primary", id="cancel")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle button presses."""
        self.dismiss(event.button.id)

    def on_key(self, event: events.Key) -> None:
        """Handle key presses."""
        match event.key.lower():
            case "x":
                event.stop()
                self.dismiss("delete")
            case "c" | "escape":
                event.stop()
                self.dismiss("cancel")
            case "d" if config["settings"]["use_recycle_bin"]:
                event.stop()
                self.dismiss("trash")



from os import getcwd, makedirs, path

from textual import work
from textual.content import Content
from textual.widgets import Button

from rovr.classes import IsValidFilePath, PathDoesntExist
from rovr.functions.icons import get_icon
from rovr.functions.path import normalise
from rovr.screens import ModalInput
from rovr.variables.constants import config


class NewItemButton(Button):
    ALLOW_MAXIMIZE = False

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            get_icon("general", "new")[0], classes="option", id="new", *args, **kwargs
        )

    def on_mount(self) -> None:
        if config["interface"]["tooltips"]:
            self.tooltip = "Create a new file or directory"

    @work
    async def on_button_pressed(self, event: Button.Pressed) -> None:
        if self.disabled:
            return
        response: str = await self.app.push_screen(
            ModalInput(
                border_title="Create New Item",
                border_subtitle="End with a slash (/) to create a directory",
                is_path=True,
                validators=[PathDoesntExist(), IsValidFilePath()],
            ),
            wait_for_dismiss=True,
        )
        if response == "":
            return
        location = normalise(path.join(getcwd(), response)) + (
            "/" if response.endswith("/") or response.endswith("\\") else ""
        )
        if location.endswith("/"):
            # recursive directory creation
            try:
                makedirs(location)
            except Exception as e:
                self.notify(
                    message=Content(f"Error creating directory '{response}': {e}"),
                    title="New Item",
                    severity="error",
                )
        elif len(location.split("/")) > 1:
            # recursive directory until file creation
            location_parts = location.split("/")
            dir_path = "/".join(location_parts[:-1])
            try:
                makedirs(dir_path)
                with open(location, "w") as f:
                    f.write("")  # Create an empty file
            except FileExistsError:
                with open(location, "w") as f:
                    f.write("")
            except Exception as e:
                self.notify(
                    message=Content(f"Error creating file '{location}': {e}"),
                    title="New Item",
                    severity="error",
                )
        else:
            # normal file creation I hope
            try:
                with open(location, "w") as f:
                    f.write("")  # Create an empty file
            except Exception as e:
                self.notify(
                    message=Content(f"Error creating file '{location}': {e}"),
                    title="New Item",
                    severity="error",
                )
        filelist = self.app.query_one("#file_list")
        await filelist.on_option_list_option_highlighted(
            filelist.OptionHighlighted(
                filelist,
                filelist.get_option_at_index(filelist.highlighted),
                filelist.highlighted,
            )
        )
        filelist.focus()



# What is textual reactive?
class SessionManager:
    """Manages session-related variables.

    Attributes:
        directories (list[dict]): A list of dictionaries that contain a
            directory's name within. The closer it is to index 0, the
            older it is.
        historyIndex (int): The index of the session in the directories.
            This can be a number between 0 and the length of the list - 1,
            inclusive.
        lastHighlighted (dict[str, int]): A dictionary mapping directory paths
            to the index of the last highlighted item. If a directory is not
            in the dictionary, the default is 0.
        selectMode (bool): Whether select mode is enabled for that directory.
        selectedItems (list[str]): A dictionary mapping directory paths to the
            list of selected items in that directory.
        search (str): The current search string.
    """

    def __init__(self) -> None:
        self.directories: list[dict] = []
        self.historyIndex: int = 0
        self.lastHighlighted: dict[str, int] = {}
        self.selectMode: bool = False
        self.selectedItems: list[str] = []
        self.search: str = ""



from typing import ClassVar, cast

from textual import events
from textual.app import ComposeResult
from textual.binding import BindingType
from textual.containers import VerticalGroup
from textual.screen import ModalScreen
from textual.widgets import OptionList

from rovr.classes.textual_options import KeybindOption
from rovr.functions import icons
from rovr.search_container import SearchInput
from rovr.variables.constants import config, schema, vindings


class KeybindList(OptionList, inherit_bindings=False):
    BINDINGS: ClassVar[list[BindingType]] = list(vindings)

    def __init__(self, **kwargs) -> None:
        keybind_data, primary_keybind_data = self.get_keybind_data()

        max_key_width = max(len(keys) for keys, _ in keybind_data)

        self.list_of_options = []
        for (keys, description), primary_key in zip(keybind_data, primary_keybind_data):
            self.list_of_options.append(
                KeybindOption(keys, description, max_key_width, primary_key)
            )
        super().__init__(*self.list_of_options, **kwargs)

    def get_keybind_data(self) -> tuple[list[tuple[str, str]], list[str]]:
        # Generate keybind data programmatically
        keybind_data = []
        primary_keys = []
        keybinds_schema = schema["properties"]["keybinds"]["properties"]
        for action, keys in config["keybinds"].items():
            if action in keybinds_schema:
                display_name = keybinds_schema[action].get("display_name", action)
                if not keys:
                    formatted_keys = "<disabled>"
                    primary_keys.append("")
                else:
                    formatted_keys = ", ".join(f"<{key}>" for key in keys)
                    primary_keys.append(keys[0])
                keybind_data.append((formatted_keys, display_name))

        # for plugins
        plugins_schema = schema["properties"]["plugins"]["properties"]
        for key, value in config["plugins"].items():
            if "enabled" in value and "keybinds" in value and key in plugins_schema:
                if not value["keybinds"] or not value["enabled"]:
                    formatted_keys = "<disabled>"
                    primary_keys.append("")
                else:
                    formatted_keys = ", ".join(f"<{key}>" for key in value["keybinds"])
                    primary_keys.append(value["keybinds"][0])
                plugins_properties = plugins_schema[key]["properties"]
                display_name = plugins_properties["keybinds"].get("display_name", key)
                keybind_data.append((formatted_keys, display_name))

        return keybind_data, primary_keys


class Keybinds(ModalScreen):
    def compose(self) -> ComposeResult:
        with VerticalGroup(id="keybinds_group"):
            yield SearchInput(
                always_add_disabled=False,
                placeholder=f"{icons.get_icon('general', 'search')[0]} Search keybinds...",
            )
            yield KeybindList(id="keybinds_data")

    def on_mount(self) -> None:
        self.input = self.query_one(SearchInput)
        self.container = self.query_one("#keybinds_group")
        self.keybinds_list = self.query_one("#keybinds_data")

        # Prevent the first focus to go to search bar
        self.keybinds_list.focus()

        self.container.border_title = "Keybinds"

        keybind_keys = config["keybinds"]["show_keybinds"]
        additional_key_string = ""
        if keybind_keys:
            short_key = "?" if keybind_keys[0] == "question_mark" else keybind_keys[0]
            additional_key_string = f"or {short_key} "
        self.container.border_subtitle = f"Press Esc {additional_key_string}to close"

    def on_key(self, event: events.Key) -> None:
        match event.key:
            case key if key in config["keybinds"]["focus_search"]:
                event.stop()
                self.input.focus()
            case key if key in config["keybinds"]["show_keybinds"] or key == "escape":
                event.stop()
                self.dismiss()

    def on_option_list_option_selected(self, event: OptionList.OptionSelected) -> None:
        if hasattr(event.option, "key_press"):
            event.stop()
            self.dismiss()
            self.app.simulate_key(cast(KeybindOption, event.option).key_press)
        else:
            raise RuntimeError(
                f"Expected a <KeybindOption> but received <{type(event.option).__name__}>"
            )



from .buttons import BackButton, ForwardButton, UpButton
from .path_input import PathAutoCompleteInput, PathInput

__all__ = [
    "PathAutoCompleteInput",
    "PathInput",
    "BackButton",
    "ForwardButton",
    "UpButton",
]



from os import getcwd, path

from rich.style import Style
from textual import on
from textual.app import ComposeResult, RenderResult
from textual.containers import Container, Horizontal, Vertical
from textual.renderables.bar import Bar as BarRenderable
from textual.widgets import Button, Input, SelectionList, Tabs
from textual.widgets._tabs import Tab, Underline
from textual.widgets.option_list import OptionDoesNotExist

from rovr.classes import SessionManager
from rovr.functions.path import normalise


class BetterBarRenderable(BarRenderable):
    HALF_BAR_LEFT: str = "╶"
    BAR: str = "─"
    HALF_BAR_RIGHT: str = "╴"


class BetterUnderline(Underline):
    def render(self) -> RenderResult:
        """Render the bar.
        Returns:
            RenderResult: the result of the render method"""
        bar_style = self.get_component_rich_style("underline--bar")
        return BetterBarRenderable(
            highlight_range=self._highlight_range,
            highlight_style=Style.from_color(bar_style.color),
            background_style=Style.from_color(bar_style.bgcolor),
        )


class TablineTab(Tab):
    def __init__(
        self, directory: str | bytes = "", label: str = "", *args, **kwargs
    ) -> None:
        """Initialise a Tab.

        Args:
            directory (str): The directory to set the tab as.
            label (ContentText): The label to use in the tab.
            id (str | None): Optional ID for the widget.
            classes (str | None): Space separated list of class names.
            disabled (bool): Whether the tab is disabled or not.
        """
        if directory == "":
            directory = getcwd()
        directory = normalise(directory)
        if label == "":
            label = str(
                path.basename(directory)
                if path.basename(directory) != ""
                else directory.strip("/")
            )
        super().__init__(label=label, *args, **kwargs)
        self.directory = directory
        self.session = SessionManager()


class Tabline(Tabs):
    def compose(self) -> ComposeResult:
        with Container(id="tabs-scroll"), Vertical(id="tabs-list-bar"):
            with Horizontal(id="tabs-list"):
                yield from self._tabs
            yield BetterUnderline()

    async def add_tab(
        self, directory: str = "", label: str = "", *args, **kwargs
    ) -> None:
        """Add a new tab to the end of the tab list.

        Args:
            directory (str): The directory to set the tab as.
            label (ContentText): The label to use in the tab.
            before (Tab | str | None): Optional tab or tab ID to add the tab before.
            after (Tab | str | None): Optional tab or tab ID to add the tab after.
        Note:
            Only one of `before` or `after` can be provided. If both are
            provided a `Tabs.TabError` will be raised.
        """
        """
        Returns:
            An optionally awaitable object that waits for the tab to be mounted and
                internal state to be fully updated to reflect the new tab.

        Raises:
            Tabs.TabError: If there is a problem with the addition request.
        """

        tab = TablineTab(directory=directory, label=label)
        super().add_tab(tab, *args, **kwargs)
        self._activate_tab(tab)
        # redo max-width
        self.parent.on_resize()

    async def remove_tab(self, tab_or_id: Tab | str | None) -> None:
        """Remove a tab.

        Args:
            tab_or_id: The Tab to remove or its id.
        """
        """
        Returns:
            An optionally awaitable object that waits for the tab to be mounted and
                internal state to be fully updated to reflect the new tab.

        Raises:
            Tabs.TabError: If there is a problem with the addition request.
        """
        super().remove_tab(tab_or_id=tab_or_id)
        self.parent.on_resize()

    @on(Tab.Clicked)
    @on(Tabs.TabActivated)
    async def check_tab_click(self, event: TablineTab.Clicked) -> None:
        assert isinstance(event.tab, TablineTab)

        def callback() -> None:
            assert isinstance(event.tab, TablineTab)
            file_list: SelectionList = self.app.query_one("#file_list")
            assert isinstance(file_list.input, Input)
            file_list.select_mode_enabled = event.tab.session.selectMode
            if event.tab.session.selectMode:
                for option in event.tab.session.selectedItems:
                    try:
                        file_list.select(file_list.get_option(option))
                        print(f"Successfully selected option: {option}")
                    except (OptionDoesNotExist, AttributeError) as e:
                        print(f"Failed to select option {option}: {e}")
            if event.tab.session.search != "":
                file_list.input.value = event.tab.session.search

        self.app.cd(event.tab.directory, add_to_history=False, callback=callback)


class NewTabButton(Button):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(label="+", variant="primary", compact=True, *args, **kwargs)

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        await self.parent.parent.query_one(Tabline).add_tab(getcwd())



from textual import events, on
from textual.app import ComposeResult
from textual.containers import Container, Grid
from textual.screen import ModalScreen
from textual.widgets import Button, Label


class Dismissable(ModalScreen):
    """Super simple screen that can be dismissed."""

    def __init__(self, message: str, border_subtitle: str = "", **kwargs) -> None:
        super().__init__(**kwargs)
        self.message = message
        self.border_subtitle = border_subtitle

    def compose(self) -> ComposeResult:
        with Grid(id="dialog"):
            yield Label(self.message, id="message")
            with Container():
                yield Button("Ok", variant="primary", id="ok")

    def on_mount(self) -> None:
        self.query_one("#ok").focus()
        self.query_one("#dialog").border_subtitle = self.border_subtitle

    def on_key(self, event: events.Key) -> None:
        """Handle key presses."""
        match event.key:
            case "escape" | "enter":
                event.stop()
                self.dismiss()
            case "tab":
                event.stop()
                self.focus_next()
            case "shift+tab":
                event.stop()
                self.focus_previous()

    @on(Button.Pressed, "#ok")
    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle button presses."""
        self.dismiss()



import platform
import shutil
import tarfile
import time
import zipfile
from contextlib import suppress
from os import getcwd, listdir, makedirs, path, remove, walk

from send2trash import send2trash
from textual import events, work
from textual.color import Gradient
from textual.containers import HorizontalGroup, VerticalGroup, VerticalScroll
from textual.renderables.bar import Bar as BarRenderable
from textual.types import UnusedParameter
from textual.widgets import Label, ProgressBar
from textual.widgets.option_list import OptionDoesNotExist

from rovr.classes import Archive
from rovr.functions import icons as icon_utils
from rovr.functions import path as path_utils
from rovr.screens import (
    CommonFileNameDoWhat,
    Dismissable,
    FileInUse,
    GiveMePermission,
    YesOrNo,
)
from rovr.variables.constants import config


class ThickBar(BarRenderable):
    HALF_BAR_LEFT = "▐"
    BAR = "█"
    HALF_BAR_RIGHT = "▌"


class ProgressBarContainer(VerticalGroup):
    def __init__(
        self,
        total: int | None = None,
        label: str = "",
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        if hasattr(self.app.get_theme(self.app.theme), "bar_gradient"):
            gradient = Gradient.from_colors(
                *self.app.get_theme(self.app.theme).bar_gradient["default"]
            )
        else:
            gradient = None
        self.progress_bar = ProgressBar(
            total=total,
            show_percentage=config["interface"]["show_progress_percentage"],
            show_eta=config["interface"]["show_progress_eta"],
            gradient=gradient,
        )
        self.progress_bar.BAR_RENDERABLE = ThickBar
        self.icon_label = Label(id="icon")
        self.text_label = Label(label, id="label")
        self.label_container = HorizontalGroup(self.icon_label, self.text_label)

    async def on_mount(self) -> None:
        await self.mount_all([self.label_container, self.progress_bar])

    def update_text(self, label: str, is_path: bool = True) -> None:
        """
        Updates the text label
        Args:
            label (str): The new label
            is_path (bool) = True: Whether the text is a path or not
        """
        if is_path and config["interface"]["truncate_progress_file_path"]:
            new_label = label.split("/")
            new_path = new_label[0]
            for path_item in new_label[1:-1]:
                new_path += "/\u2026"
            new_path += f"/{new_label[-1]}"
            label = new_path
        self.text_label.update(label)

    def update_icon(self, icon: str) -> None:
        """
        Updates the icon label
        Args:
            icon (str): The new icon
        """
        self.icon_label.update(icon)

    def update_progress(
        self,
        total: None | float | UnusedParameter = UnusedParameter(),
        progress: float | UnusedParameter = UnusedParameter(),
        advance: float | UnusedParameter = UnusedParameter(),
    ) -> None:
        self.progress_bar.update(total=total, progress=progress, advance=advance)

    def panic(
        self,
        dismiss_with: dict | None = None,
        notify: dict | None = None,
        bar_text: str = "",
    ) -> None:
        """Do something when an error occurs.
        Args:
            dismiss_with(dict): The message for the Dismissable screen (must contain `message` and `subtitle`)
            notify(dict): The notify message (must contain `message` and `title`)
            bar_text(str): The new text to update the label
        """
        if bar_text:
            self.update_text(bar_text, False)
        if self.progress_bar.total is None:
            self.progress_bar.update(total=1, progress=0)
        self.add_class("error")
        if hasattr(self.app.get_theme(self.app.theme), "bar_gradient"):
            self.progress_bar.gradient = Gradient.from_colors(
                *self.app.get_theme(self.app.theme).bar_gradient["error"]
            )
        self.update_icon(
            self.icon_label.content + " " + icon_utils.get_icon("general", "close")[0]
        )
        dismiss_with = dismiss_with or {}
        notify = notify or {}

        if dismiss_with:
            self.app.call_from_thread(
                self.app.push_screen_wait,
                Dismissable(
                    dismiss_with["message"], border_subtitle=dismiss_with["subtitle"]
                ),
            )
        if notify:
            self.notify(
                message=notify["message"], severity="error", title=notify["title"]
            )


class ProcessContainer(VerticalScroll):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(id="processes", *args, **kwargs)

    async def new_process_bar(
        self, max: int | None = None, id: str | None = None, classes: str | None = None
    ) -> ProgressBarContainer:
        new_bar: ProgressBarContainer = ProgressBarContainer(
            total=max, id=id, classes=classes
        )
        await self.mount(new_bar, before=0)
        return new_bar

    @work(thread=True)
    def delete_files(
        self, files: list[str], compressed: bool = True, ignore_trash: bool = False
    ) -> None:
        """
        Remove files from the filesystem.

        Args:
            files (list[str]): List of file paths to remove.
            compressed (bool): Whether the file paths are compressed. Defaults to True.
            ignore_trash (bool): If True, files will be permanently deleted instead of sent to the recycle bin. Defaults to False.
        """
        # Create progress/process bar (why have I set names as such...)
        bar: ProgressBarContainer = self.app.call_from_thread(
            self.new_process_bar, classes="active"
        )
        self.app.call_from_thread(
            bar.update_icon,
            icon_utils.get_icon("general", "delete")[0],
        )
        self.app.call_from_thread(
            bar.update_text,
            "Getting files to delete...",
        )
        # get files to delete
        files_to_delete = []
        folders_to_delete = []
        for file in files:
            if compressed:
                file = path_utils.decompress(file)
            if path.isdir(file):
                folders_to_delete.append(file)
            files_to_add, folders_to_add = path_utils.get_recursive_files(
                file, with_folders=True
            )
            files_to_delete.extend(files_to_add)
            folders_to_delete.extend(folders_to_add)
        self.app.call_from_thread(bar.update_progress, total=len(files_to_delete) + 1)
        action_on_permission_error = "ask"
        last_update_time = time.monotonic()
        for i, item_dict in enumerate(files_to_delete):
            current_time = time.monotonic()
            if (
                current_time - last_update_time > 0.25
                or i == len(files_to_delete) - 1
                or i == 0
            ):
                self.app.call_from_thread(
                    bar.update_text,
                    item_dict["relative_loc"],
                )
                self.app.call_from_thread(bar.update_progress, progress=i + 1)
                last_update_time = current_time
            if path.exists(item_dict["path"]):
                # I know that it `path.exists` prevents issues, but on the
                # off chance that anything happens, this should help
                try:
                    if config["settings"]["use_recycle_bin"] and not ignore_trash:
                        try:
                            path_to_trash = item_dict["path"]
                            if platform.system() == "Windows":
                                # An inherent issue with long paths on windows
                                path_to_trash = path_to_trash.replace("/", "\\")
                                pass
                            send2trash(path_to_trash)
                        except PermissionError as e:
                            # On Windows, a file being used by another process
                            # raises a PermissionError/OSError with winerror 32.
                            is_file_in_use = False
                            try:
                                # Some PermissionError instances expose winerror
                                is_file_in_use = getattr(e, "winerror", None) == 32
                            except Exception:
                                is_file_in_use = False
                            if is_file_in_use and platform.system() == "Windows":
                                # Ask the user specifically that the file is in use
                                response = self.app.call_from_thread(
                                    self.app.push_screen_wait,
                                    FileInUse(
                                        f"The file appears to be open in another application and cannot be deleted.\nPath: {item_dict['relative_loc']}",
                                    ),
                                )
                                if not response["value"]:
                                    bar.panic()
                                    return
                                else:
                                    continue
                            elif is_file_in_use:
                                # need to ensure unix users see an
                                # error so they create an issue
                                self.app.panic()
                            # fallback for regular permission issues
                            if action_on_permission_error == "ask":
                                do_what = self.app.call_from_thread(
                                    self.app.push_screen_wait,
                                    GiveMePermission(
                                        "Path has no write access to be deleted.\nForcefully obtain and delete it?",
                                        border_title=item_dict["path"],
                                    ),
                                )
                                if do_what["toggle"]:
                                    action_on_permission_error = do_what["value"]
                                action = do_what["value"]
                            else:
                                action = action_on_permission_error
                            match action:
                                case "force":
                                    if path_utils.force_obtain_write_permission(
                                        item_dict["path"]
                                    ):
                                        remove(item_dict["path"])
                                case "skip":
                                    continue
                                case "cancel":
                                    bar.panic()
                                    return
                        except Exception as e:
                            do_what = self.app.call_from_thread(
                                self.app.push_screen_wait,
                                YesOrNo(
                                    f"Trashing failed due to\n{e}\nDo Permenant Deletion?",
                                    with_toggle=True,
                                    border_subtitle="If this is a bug, please file an issue!",
                                ),
                            )
                            if do_what["toggle"]:
                                ignore_trash = do_what["value"]
                            if do_what["value"]:
                                remove(item_dict["path"])
                            else:
                                continue
                    else:
                        remove(item_dict["path"])
                except FileNotFoundError:
                    # it's deleted, so why care?
                    pass
                except PermissionError as e:
                    # Try to detect if file is in use on Windows
                    is_file_in_use = False
                    try:
                        is_file_in_use = getattr(e, "winerror", None) == 32
                    except Exception:
                        is_file_in_use = False
                    if is_file_in_use and platform.system() == "Windows":
                        response = self.app.call_from_thread(
                            self.app.push_screen_wait,
                            FileInUse(
                                f"The file appears to be open in another application and cannot be deleted.\nPath: {item_dict['relative_loc']}",
                            ),
                        )
                        if not response["value"]:
                            bar.panic()
                            return
                        else:
                            continue
                    elif is_file_in_use:
                        # need to ensure unix users see an
                        # error so they create an issue
                        self.app.panic()
                    # fallback for regular permission issues
                    if action_on_permission_error == "ask":
                        do_what = self.app.call_from_thread(
                            self.app.push_screen_wait,
                            GiveMePermission(
                                "Path has no write access to be deleted.\nForcefully obtain and delete it?",
                                border_title=item_dict["path"],
                            ),
                        )
                        if do_what["toggle"]:
                            action_on_permission_error = do_what["value"]
                        action = do_what["value"]
                    else:
                        action = action_on_permission_error
                    match action:
                        case "force":
                            if path_utils.force_obtain_write_permission(
                                item_dict["path"]
                            ):
                                remove(item_dict["path"])
                        case "skip":
                            continue
                        case "cancel":
                            bar.panic()
                            return
                except Exception as e:
                    # TODO: should probably let it continue, then have a summary
                    bar.panic(
                        dismiss_with={
                            "message": f"Deleting failed due to\n{e}\nProcess Aborted.",
                            "subtitle": "If this is a bug, please file an issue!",
                        },
                        bar_text="Unhandled Error",
                    )
                    return
        # The reason for an extra +1 in the total is for this
        # handling folders
        has_perm_error = False
        for folder in folders_to_delete:
            try:
                shutil.rmtree(folder)
            except PermissionError:
                has_perm_error = True
            except FileNotFoundError:
                # ig it got removed?
                pass
        if has_perm_error:
            bar.panic(
                notify={
                    "message": f"Certain files in {folder} could not be deleted due to PermissionError.",
                    "title": "Delete Files",
                },
            )
            return
        # if there werent any files, show something useful
        # aside from 'Getting files to delete...'
        if files_to_delete == [] and folders_to_delete != []:
            self.app.call_from_thread(
                bar.update_text,
                folders_to_delete[-1],
            )
        elif files_to_delete == folders_to_delete == []:
            # this cannot happen, but just as an easter egg :shippit:
            self.app.call_from_thread(
                bar.update_text, "Successfully deleted nothing!", False
            )
        # finished successfully
        self.app.call_from_thread(
            bar.update_icon,
            bar.icon_label.content + " " + icon_utils.get_icon("general", "check")[0],
        )
        self.app.call_from_thread(bar.progress_bar.advance)
        self.app.call_from_thread(bar.add_class, "done")

    @work(thread=True)
    def create_archive(self, files: list[str], archive_name: str) -> None:
        """
        Compress files into an archive.

        Args:
            files (list[str]): List of file paths to compress.
            archive_name (str): Path for the output archive.
        """
        bar: ProgressBarContainer = self.app.call_from_thread(
            self.new_process_bar, classes="active"
        )
        self.app.call_from_thread(
            bar.update_icon,
            icon_utils.get_icon("general", "zip")[0],
        )
        self.app.call_from_thread(bar.update_text, "Getting files to archive...", False)

        files_to_archive = []
        for p in files:
            if path.isdir(p):
                if not listdir(p):  # empty directory
                    files_to_archive.append(p)
                else:
                    for dirpath, _, filenames in walk(p):
                        for f in filenames:
                            files_to_archive.append(path.join(dirpath, f))
            else:
                files_to_archive.append(p)

        files_to_archive = sorted(list(set(files_to_archive)))

        self.app.call_from_thread(bar.update_progress, total=len(files_to_archive) + 1)

        if len(files) == 1:
            base_path = path.dirname(files[0])
        else:
            base_path = path.commonpath(files)

        try:
            with Archive(archive_name, "w") as archive:
                assert archive._archive is not None
                last_update_time = time.monotonic()
                for i, file_path in enumerate(files_to_archive):
                    arcname = path.relpath(file_path, base_path)
                    current_time = time.monotonic()
                    if (
                        current_time - last_update_time > 0.25
                        or i == len(files_to_archive) - 1
                    ):
                        self.app.call_from_thread(
                            bar.update_text,
                            arcname,
                        )
                        self.app.call_from_thread(bar.update_progress, progress=i + 1)
                        last_update_time = current_time
                    _archive = archive._archive
                    if _archive:
                        if archive._is_zip:
                            assert isinstance(_archive, zipfile.ZipFile)
                            _archive.write(file_path, arcname=arcname)
                        else:
                            assert isinstance(_archive, tarfile.TarFile)
                            _archive.add(file_path, arcname=arcname)
                for p in files:
                    if path.isdir(p) and not listdir(p):
                        arcname = path.relpath(p, base_path)
                        _archive = archive._archive
                        if _archive:
                            if archive._is_zip:
                                assert isinstance(_archive, zipfile.ZipFile)
                                _archive.write(p, arcname=arcname)
                            else:
                                assert isinstance(_archive, tarfile.TarFile)
                                _archive.add(p, arcname=arcname)

        except Exception as e:
            bar.panic(
                dismiss_with={
                    "message": f"Archiving failed due to\n{e}\nProcess Aborted.",
                    "subtitle": "File an issue if this is a bug!",
                }
            )
            return

        self.app.call_from_thread(
            bar.update_icon,
            bar.icon_label.content + " " + icon_utils.get_icon("general", "check")[0],
        )
        self.app.call_from_thread(bar.progress_bar.advance)
        self.app.call_from_thread(bar.add_class, "done")

    @work(thread=True)
    def unzip_file(self, archive_path: str, destination_path: str) -> None:
        """
        Extracts a zip archive to a destination.

        Args:
            archive_path (str): Path to the zip archive.
            destination_path (str): Path to the destination folder.
        """
        bar: ProgressBarContainer = self.app.call_from_thread(
            self.new_process_bar, classes="active"
        )
        self.app.call_from_thread(
            bar.update_icon,
            icon_utils.get_icon("general", "open")[0],
        )
        self.app.call_from_thread(
            bar.update_text,
            "Preparing to extract...",
        )

        do_what_on_existance = "ask"
        action_on_permission_error = "ask"
        try:
            if not path.exists(destination_path):
                makedirs(destination_path)

            with Archive(archive_path, "r") as archive:
                file_list = archive.infolist()
                self.app.call_from_thread(bar.update_progress, total=len(file_list) + 1)

                last_update_time = time.monotonic()
                for i, file in enumerate(file_list):
                    filename = getattr(file, "filename", getattr(file, "name", ""))
                    current_time = time.monotonic()
                    if (
                        current_time - last_update_time > 0.25
                        or i == len(file_list) - 1
                        or i == 0
                    ):
                        self.app.call_from_thread(
                            bar.update_text,
                            filename,
                        )
                        self.app.call_from_thread(bar.update_progress, progress=i + 1)
                        last_update_time = current_time
                    if path.exists(path.join(destination_path, filename)):
                        if do_what_on_existance == "ask":
                            response = self.app.call_from_thread(
                                self.app.push_screen_wait,
                                CommonFileNameDoWhat(
                                    "Path already exists in destination\nWhat do you want to do now?",
                                    border_title=filename,
                                    border_subtitle=f"Extracting to {destination_path}",
                                ),
                            )
                            if response["same_for_next"]:
                                do_what_on_existance = response["value"]
                            val = response["value"]
                        else:
                            val = do_what_on_existance
                        match val:
                            case "overwrite":
                                pass
                            case "skip":
                                continue
                            case "rename":
                                base_name, extension = path.splitext(filename)
                                tested_number = 1
                                while True:
                                    new_filename = (
                                        f"{base_name} ({tested_number}){extension}"
                                    )
                                    new_path = path_utils.normalise(
                                        path.join(destination_path, new_filename)
                                    )
                                    if not path.exists(new_path):
                                        break
                                    tested_number += 1

                                source = archive.open(file)
                                if source:
                                    with source, open(new_path, "wb") as target:
                                        shutil.copyfileobj(source, target)
                                continue
                            case "cancel":
                                bar.panic()
                                return
                    try:
                        archive.extract(file, path=destination_path)
                    except PermissionError:
                        if action_on_permission_error == "ask":
                            do_what = self.app.call_from_thread(
                                self.app.push_screen_wait,
                                GiveMePermission(
                                    "Path has no write access to be overwritten.\nForcefully obtain and overwrite?",
                                    border_title=filename,
                                ),
                            )
                            if do_what["toggle"]:
                                action_on_permission_error = do_what["value"]
                            action = do_what["value"]
                        else:
                            action = action_on_permission_error
                        match action:
                            case "force":
                                if path_utils.force_obtain_write_permission(
                                    path.join(destination_path, filename)
                                ):
                                    archive.extract(file, path=destination_path)
                            case "skip":
                                continue
                            case "cancel":
                                bar.panic()
                                return
        except (zipfile.BadZipFile, tarfile.TarError, ValueError) as e:
            dismiss_with = {"subtitle": ""}
            if isinstance(e, ValueError) and "Password" in e.__str__():
                if "ZIP" in e.__str__():
                    dismiss_with["message"] = (
                        "Password-protected ZIP files cannot be unzipped"
                    )
                elif "RAR" in e.__str__():
                    dismiss_with["message"] = (
                        "Password-protected RAR files cannot be unzipped"
                    )
                else:
                    dismiss_with["message"] = (
                        "Password-protected archive files cannot be unzipped"
                    )
            else:
                dismiss_with = {
                    "message": f"Unzipping failed due to {type(e).__name__}\n{e}\nProcess Aborted.",
                    "subtitle": "If this is a bug, file an issue!",
                }
            bar.panic(dismiss_with=dismiss_with, bar_text="Error extracting archive")
            return
        except Exception as e:
            bar.panic(
                dismiss_with={
                    "message": f"Unzipping failed due to {type(e).__name__}\n{e}\nProcess Aborted.",
                    "subtitle": "If this is a bug, please file an issue!",
                },
                bar_text="Unhandled Error",
            )
            return

        self.app.call_from_thread(
            bar.update_icon,
            icon_utils.get_icon("general", "check")[0],
        )
        self.app.call_from_thread(bar.progress_bar.advance)
        self.app.call_from_thread(bar.add_class, "done")

    @work(thread=True)
    def paste_items(self, copied: list[str], cutted: list[str], dest: str = "") -> None:
        """
        Paste copied or cut files to the current directory
        Args:
            copied (list[str]): A list of items to be copied to the location
            cutted (list[str]): A list of items to be cut to the location
            dest (str) = getcwd(): The directory to copy to.
        """
        if dest == "":
            dest = getcwd()
        bar: ProgressBarContainer = self.app.call_from_thread(
            self.new_process_bar, classes="active"
        )
        self.app.call_from_thread(
            bar.update_icon,
            icon_utils.get_icon("general", "paste")[0],
        )
        self.app.call_from_thread(
            bar.update_text,
            "Getting items to paste...",
        )
        files_to_copy = []
        files_to_cut = []
        cut_files__folders = []
        for file in copied:
            files_to_copy.extend(path_utils.get_recursive_files(file))
        for file in cutted:
            if path.isdir(file):
                cut_files__folders.append(path_utils.normalise(file))
            files, folders = path_utils.get_recursive_files(file, with_folders=True)
            files_to_cut.extend(files)
            cut_files__folders.extend(folders)
        self.app.call_from_thread(
            bar.update_progress, total=int(len(files_to_copy) + len(files_to_cut)) + 1
        )
        action_on_existance = "ask"
        action_on_permission_error = "ask"
        last_update_time = time.monotonic()
        if files_to_copy:
            self.app.call_from_thread(
                bar.update_icon,
                icon_utils.get_icon("general", "copy")[0],
            )
        for i, item_dict in enumerate(files_to_copy):
            current_time = time.monotonic()
            if (
                current_time - last_update_time > 0.25
                or i == len(files_to_copy) - 1
                or i == 0
            ):
                self.app.call_from_thread(
                    bar.update_text,
                    item_dict["relative_loc"],
                )
                last_update_time = current_time
                self.app.call_from_thread(bar.update_progress, progress=i + 1)
            if path.exists(item_dict["path"]):
                # again checks just in case something goes wrong
                try:
                    makedirs(
                        path_utils.normalise(
                            path.join(dest, item_dict["relative_loc"], "..")
                        ),
                        exist_ok=True,
                    )
                    if path.exists(path.join(dest, item_dict["relative_loc"])):
                        # check if overwrite
                        if action_on_existance == "ask":
                            response = self.app.call_from_thread(
                                self.app.push_screen_wait,
                                CommonFileNameDoWhat(
                                    "The destination already has file of that name.\nWhat do you want to do now?",
                                    border_title=item_dict["relative_loc"],
                                    border_subtitle=f"Copying to {dest}",
                                ),
                            )
                            if response["same_for_next"]:
                                action_on_existance = response["value"]
                            val = response["value"]
                        else:
                            val = action_on_existance
                        match val:
                            case "overwrite":
                                pass
                            case "skip":
                                continue
                            case "rename":
                                base_name, extension = path.splitext(
                                    item_dict["relative_loc"]
                                )
                                tested_number = 1
                                while True:
                                    new_rel_path = (
                                        f"{base_name} ({tested_number}){extension}"
                                    )
                                    if not path.exists(path.join(dest, new_rel_path)):
                                        break
                                    tested_number += 1
                                item_dict["relative_loc"] = new_rel_path
                            case "cancel":
                                bar.panic(bar_text="Process cancelled.")
                                return
                    if config["settings"]["copy_includes_metadata"]:
                        shutil.copy2(
                            item_dict["path"],
                            path.join(dest, item_dict["relative_loc"]),
                        )
                    else:
                        shutil.copy(
                            item_dict["path"],
                            path.join(dest, item_dict["relative_loc"]),
                        )
                except (OSError, PermissionError):
                    # OSError from shutil: The destination location must be writable;
                    # otherwise, an OSError exception will be raised
                    # Permission Error just in case
                    if action_on_permission_error == "ask":
                        do_what = self.app.call_from_thread(
                            self.app.push_screen_wait,
                            GiveMePermission(
                                "Path has no write access to be overwritten.\nForefully obtain and overwrite?",
                                border_title=item_dict["relative_loc"],
                            ),
                        )
                        if do_what["toggle"]:
                            action_on_permission_error = do_what["value"]
                        action = do_what["value"]
                    else:
                        action = action_on_permission_error
                    match action:
                        case "force":
                            if path_utils.force_obtain_write_permission(
                                path.join(dest, item_dict["relative_loc"])
                            ):
                                shutil.copy(
                                    item_dict["path"],
                                    path.join(dest, item_dict["relative_loc"]),
                                )
                        case "skip":
                            continue
                        case "cancel":
                            self.app.call_from_thread(bar.add_class, "error")
                            self.app.call_from_thread(
                                bar.update_icon,
                                icon_utils.get_icon("general", "copy")[0]
                                + " "
                                + icon_utils.get_icon("general", "close")[0],
                            )
                            return
                except FileNotFoundError:
                    # the only way this can happen is if the file is deleted
                    # midway through the process, which means the user is
                    # literally testing the limits, so yeah uhh, pass
                    pass
                except Exception as e:
                    # TODO: should probably let it continue, then have a summary
                    bar.panic(
                        dismiss_with={
                            "message": f"Copying failed due to {type(e).__name__}\n{e}\nProcess Aborted.",
                            "subtitle": "If this is a bug, please file an issue!",
                        },
                        bar_text="Unhandled Error",
                    )
                    return

        cut_ignore = []
        action_on_permission_error = "ask"
        last_update_time = time.monotonic()
        if files_to_cut:
            self.app.call_from_thread(
                bar.update_icon,
                icon_utils.get_icon("general", "cut")[0],
            )
        for i, item_dict in enumerate(files_to_cut):
            current_time = time.monotonic()
            if (
                current_time - last_update_time > 0.25
                or i == len(files_to_cut) - 1
                or i == 0
            ):
                self.app.call_from_thread(
                    bar.update_text,
                    item_dict["relative_loc"],
                )
                self.app.call_from_thread(bar.update_progress, progress=i + 1)
                last_update_time = current_time
            if path.exists(item_dict["path"]):
                # again checks just in case something goes wrong
                try:
                    makedirs(
                        path_utils.normalise(
                            path.join(dest, item_dict["relative_loc"], "..")
                        ),
                        exist_ok=True,
                    )
                    if path.exists(path.join(dest, item_dict["relative_loc"])):
                        print(
                            path_utils.normalise(
                                path.join(dest, item_dict["relative_loc"])
                            ),
                            path_utils.normalise(item_dict["path"]),
                        )

                        if path_utils.normalise(
                            path.join(dest, item_dict["relative_loc"])
                        ) == path_utils.normalise(item_dict["path"]):
                            cut_ignore.append(item_dict["path"])
                            continue
                        if action_on_existance == "ask":
                            response = self.app.call_from_thread(
                                self.app.push_screen_wait,
                                CommonFileNameDoWhat(
                                    "The destination already has file of that name.\nWhat do you want to do now?",
                                    border_title=item_dict["relative_loc"],
                                    border_subtitle=f"Moving to {dest}",
                                ),
                            )
                            if response["same_for_next"]:
                                action_on_existance = response["value"]
                            val = response["value"]
                        else:
                            val = action_on_existance
                        match val:
                            case "overwrite":
                                pass
                            case "skip":
                                cut_ignore.append(item_dict["path"])
                                continue
                            case "rename":
                                base_name, extension = path.splitext(
                                    item_dict["relative_loc"]
                                )
                                tested_number = 1
                                while True:
                                    new_rel_path = (
                                        f"{base_name} ({tested_number}){extension}"
                                    )
                                    if not path.exists(path.join(dest, new_rel_path)):
                                        break
                                    tested_number += 1
                                item_dict["relative_loc"] = new_rel_path
                            case "cancel":
                                bar.panic(bar_text="Process cancelled.")
                                return
                    shutil.move(
                        item_dict["path"],
                        path.join(dest, item_dict["relative_loc"]),
                    )
                except (OSError, PermissionError):
                    # OSError from shutil: The destination location must be writable;
                    # otherwise, an OSError exception will be raised
                    # Permission Error just in case
                    if action_on_permission_error == "ask":
                        do_what = self.app.call_from_thread(
                            self.app.push_screen_wait,
                            GiveMePermission(
                                "Path has no write access to be overwritten.\nForcefully obtain and overwrite?",
                                border_title=item_dict["relative_loc"],
                            ),
                        )
                        if do_what["toggle"]:
                            action_on_permission_error = do_what["value"]
                        action = do_what["value"]
                    else:
                        action = action_on_permission_error
                    match action:
                        case "force":
                            if path_utils.force_obtain_write_permission(
                                path.join(dest, item_dict["relative_loc"])
                            ) and path_utils.force_obtain_write_permission(
                                item_dict["path"]
                            ):
                                shutil.move(
                                    item_dict["path"],
                                    path.join(dest, item_dict["relative_loc"]),
                                )
                        case "skip":
                            continue
                        case "cancel":
                            bar.panic(bar_text="Process cancelled.")
                            return
                except FileNotFoundError:
                    # the only way this can happen is if the file is deleted
                    # midway through the process, which means the user is
                    # literally testing the limits, so yeah uhh, pass
                    pass
                except Exception as e:
                    # TODO: should probably let it continue, then have a summary
                    bar.panic(
                        dismiss_with={
                            "message": f"Moving failed due to {type(e).__name__}\n{e}\nProcess Aborted.",
                            "subtitle": "If this is a bug, please file an issue!",
                        },
                        bar_text="Unhandled Error",
                    )
                    return
        # delete the folders
        has_perm_error = False
        for folder in cut_files__folders:
            try:
                skip = False
                for file in cut_ignore:
                    if folder in file:
                        skip = True
                        break
                if not skip:
                    shutil.rmtree(folder)
            except PermissionError:
                has_perm_error = True
            except FileNotFoundError:
                # ig it got removed?
                continue
        if has_perm_error:
            bar.panic(
                notify={
                    "message": f"Certain files in {folder} could not be deleted due to PermissionError.",
                    "title": "Delete Files",
                },
                bar_text=path.basename(cutted[-1]),
            )
            return
        # remove from clipboard
        for item in cutted:
            # cant bother to figure out how this happens,
            # just catch it
            with suppress(OptionDoesNotExist):
                self.app.call_from_thread(
                    self.app.query_one("Clipboard").remove_option,
                    path_utils.compress(item),
                )
        self.app.call_from_thread(
            bar.update_icon,
            icon_utils.get_icon("general", "cut" if len(cutted) else "copy")[0],
        )
        self.app.call_from_thread(
            bar.update_icon,
            icon_utils.get_icon("general", "check")[0],
        )
        self.app.call_from_thread(bar.progress_bar.advance)
        self.app.call_from_thread(bar.add_class, "done")

    async def on_key(self, event: events.Key) -> None:
        if event.key in config["keybinds"]["delete"]:
            event.stop()
            await self.remove_children(".done")
            await self.remove_children(".error")



from textual.widgets import Button

from rovr.functions.icons import get_icon
from rovr.functions.path import decompress
from rovr.screens import YesOrNo
from rovr.variables.constants import config


class PasteButton(Button):
    ALLOW_MAXIMIZE = False

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            get_icon("general", "paste")[0],
            classes="option",
            id="paste",
            *args,
            **kwargs,
        )

    def on_mount(self) -> None:
        if config["interface"]["tooltips"]:
            self.tooltip = "Paste files from clipboard"

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        if self.disabled:
            return
        """Paste files from clipboard"""
        selected_items = self.app.query_one(
            "Clipboard"
        ).selected  # dont include highlighted
        if selected_items:
            # decompress items
            selected_items = [decompress(item) for item in selected_items]
            # split into two items, those ending with `-cut` and those ending with `-copy`
            to_copy, to_cut = (
                [item[:-5] for item in selected_items if item.endswith("-copy")],
                [item[:-4] for item in selected_items if item.endswith("-cut")],
            )

            async def callback(response: str) -> None:
                """Callback to paste files after confirmation"""
                if response:
                    self.app.query_one("ProcessContainer").paste_items(to_copy, to_cut)

            self.app.push_screen(
                YesOrNo(
                    message="Are you sure you want to "
                    + (
                        f"copy {len(to_copy)} item{'s' if len(to_copy) != 1 else ''}{' and ' if len(to_cut) != 0 else ''}"
                        if len(to_copy) > 0
                        else ""
                    )
                    + (
                        f"cut {len(to_cut)} item{'s' if len(to_cut) != 1 else ''}"
                        if len(to_cut) > 0
                        else ""
                    )
                    + "?"
                ),
                callback=callback,
            )



import asyncio
import ctypes
import os
import stat
from os import path

import psutil
from lzstring import LZString
from rich.console import Console
from textual import work
from textual.app import App

from rovr.functions.icons import get_icon_for_file, get_icon_for_folder
from rovr.variables.constants import os_type

lzstring = LZString()
pprint = Console().print


config = {}
pins = {}


def normalise(location: str | bytes) -> str:
    """'Normalise' the path
    Args:
        location (str): The location to the item

    Returns:
        str: A normalised path
    """
    # path.normalise fixes the relative references
    # replace \\ with / on windows
    # by any chance if somehow a \\\\ was to enter, fix that
    return str(path.normpath(location).replace("\\", "/").replace("//", "/"))


def is_hidden_file(filepath: str) -> bool:
    if os_type == "Windows":
        try:
            GetFileAttributesW = ctypes.windll.kernel32.GetFileAttributesW
            GetFileAttributesW.argtypes = [ctypes.c_wchar_p]
            GetFileAttributesW.restype = ctypes.c_uint32
            attrs = GetFileAttributesW(filepath)
            if attrs == 0xFFFFFFFF:  # INVALID_FILE_ATTRIBUTES
                return False
            return bool(attrs & 0x02)  # FILE_ATTRIBUTE_HIDDEN
        except (OSError, AttributeError):
            return False
    elif os_type == "Darwin":
        # dotfiles should always be hidden, and so should UF_HIDDEN-flagged files
        name_hidden = path.basename(filepath).startswith(".")
        try:
            st = os.stat(filepath, follow_symlinks=False)
            flag_hidden = bool(
                getattr(st, "st_flags", 0) & getattr(stat, "UF_HIDDEN", 0)
            )
        except OSError:
            flag_hidden = False
        return name_hidden or flag_hidden
    else:
        return path.basename(filepath).startswith(".")


# Okay so the reason why I have wrapper functions is
# I was messing around with different LZString options
# and Encoded URI Component seems to best option. I've just
# left it here, in case we can switch to something like
# base 64 because Encoded URI Component can get quite long
# very fast, which isn't really the purpose of LZString
def compress(text: str) -> str:
    return lzstring.compressToEncodedURIComponent(text)


def decompress(text: str) -> str:
    return lzstring.decompressFromEncodedURIComponent(text)


@work
async def open_file(app: App, filepath: str) -> None:
    """Cross-platform function to open files with their default application.

    Args:
        app (App): The Textuall application instance
        filepath (str): Path to the file to open
    """
    system = os_type.lower()

    try:
        match system:
            case "windows":
                process = await asyncio.create_subprocess_exec(
                    "cmd",
                    "/c",
                    "start",
                    "",
                    filepath,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                )
            case "darwin":  # macOS
                process = await asyncio.create_subprocess_exec(
                    "open",
                    filepath,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                )
            case _:  # Linux and other Unix-like
                process = await asyncio.create_subprocess_exec(
                    "xdg-open",
                    filepath,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                )
        _, stderr = await process.communicate()
        if stderr:
            app.notify(
                str(stderr.decode().strip()), title="Open File", severity="error"
            )
        elif process.returncode and process.returncode != 0:
            app.notify(
                f"Process exited with return code {process.returncode}",
                title="Open File",
                severity="error",
            )
    except Exception as e:
        app.notify(str(e), title="Open File", severity="error")


def get_filtered_dir_names(cwd: str | bytes, show_hidden: bool = False) -> set[str]:
    """
    Get the names of all items in a directory, respecting the show_hidden setting.
    This function is used for comparison in file watchers to avoid refresh loops.

    Args:
        cwd(str): The working directory to check
        show_hidden(bool): Whether to include hidden files/folders

    Returns:
        set[str]: A set of item names in the directory

    Raises:
        PermissionError: When access to the directory is denied
    """
    try:
        listed_dir = os.scandir(cwd)
    except (PermissionError, FileNotFoundError, OSError):
        raise PermissionError(f"PermissionError: Unable to access {cwd}")

    names = set()
    for item in listed_dir:
        if not show_hidden and is_hidden_file(item.path):
            continue
        names.add(item.name)

    return names


def get_cwd_object(
    cwd: str, show_hidden: bool = False
) -> tuple[list[dict], list[dict]]:
    """
    Get the objects (files and folders) in a provided directory
    Args:
        cwd(str): The working directory to check
        show_hidden(bool): Whether to include hidden files/folders (dot-prefixed on Unix; flagged hidden on Windows/macOS)
    Returns:
        folders(list[dict]): A list of dictionaries, containing "name" as the item's name and "icon" as the respective icon
        files(list[dict]): A list of dictionaries, containing "name" as the item's name and "icon" as the respective icon

    Raises:
        PermissionError: When access to the directory is denied
    """
    folders, files = [], []
    try:
        listed_dir = os.scandir(cwd)
    except (PermissionError, FileNotFoundError, OSError):
        raise PermissionError(f"PermissionError: Unable to access {cwd}")
    for item in listed_dir:
        # Skip hidden files if show_hidden is False
        if not show_hidden and is_hidden_file(item.path):
            continue

        if item.is_dir():
            folders.append({
                "name": item.name,
                "icon": get_icon_for_folder(item.name),
                "dir_entry": item,
            })
        else:
            files.append({
                "name": item.name,
                "icon": get_icon_for_file(item.name),
                "dir_entry": item,
            })
    # Sort folders and files properly
    folders.sort(key=lambda x: x["name"].lower())
    files.sort(key=lambda x: x["name"].lower())
    print(f"Found {len(folders)} folders and {len(files)} files in {cwd}")
    return folders, files


def file_is_type(file_path: str) -> str:
    """Get a given path's type
    Args:
        file_path(str): The file path to check

    Returns:
        str: The string that says what type it is (unknown, symlink, directory, junction or file)
    """
    try:
        file_stat = os.lstat(file_path)
    except (OSError, FileNotFoundError):
        return "unknown"
    mode = file_stat.st_mode
    if stat.S_ISLNK(mode):
        return "symlink"
    elif stat.S_ISDIR(mode):
        return "directory"
    elif (
        os_type == "Windows"
        and hasattr(file_stat, "st_file_attributes")
        and file_stat.st_file_attributes & stat.FILE_ATTRIBUTE_REPARSE_POINT
    ):
        return "junction"
    else:
        return "file"


def force_obtain_write_permission(item_path: str) -> bool:
    """
    Forcefully obtain write permission to a file or directory.

    Args:
        item_path (str): The path to the file or directory.

    Returns:
        bool: True if permission was granted successfully, False otherwise.
    """
    if not path.exists(item_path):
        return False
    try:
        current_permissions = stat.S_IMODE(os.lstat(item_path).st_mode)
        os.chmod(item_path, current_permissions | stat.S_IWRITE)
        return True
    except (OSError, PermissionError) as e:
        pprint(
            f"[bright_red]Permission Error:[/] Failed to change permission for {item_path}: {e}"
        )
        return False


def get_recursive_files(
    object_path: str, with_folders: bool = False
) -> list[dict] | tuple[list[dict], list[dict]]:
    """Get the files available at a directory recursively, regardless of whether it is a directory or not
    Args:
        object_path (str): The object's path
        with_folders (bool): Return a list of folders as well

    Returns:
        list: A list of dictionaries, with a "path" key and "relative_loc" key
        OR
        list: A list of dictionaries, with a "path" key and "relative_loc" key for files
        list: A list of path strings that were involved in the file list.
    """
    if path.isfile(path.realpath(object_path)) or path.islink(
        path.realpath(object_path)
    ):
        if with_folders:
            return [
                {
                    "path": normalise(object_path),
                    "relative_loc": path.basename(object_path),
                }
            ], []
        return [
            {
                "path": normalise(object_path),
                "relative_loc": path.basename(object_path),
            }
        ]
    else:
        files = []
        folders = []
        for folder, folders_in_folder, files_in_folder in os.walk(object_path):
            if with_folders:
                for folder_in_folder in folders_in_folder:
                    full_path = normalise(path.join(folder, folder_in_folder))
                    if full_path not in folder:
                        folders.append(full_path)
            for file in files_in_folder:
                full_path = normalise(path.join(folder, file))  # normalise the path
                files.append({
                    "path": full_path,
                    "relative_loc": normalise(
                        path.relpath(full_path, object_path + "/..")
                    ),
                })
        if with_folders:
            return files, folders
        return files


def ensure_existing_directory(directory: str) -> str:
    while not (path.exists(directory) and path.isdir(directory)):
        parent = path.dirname(directory)
        # If we can't even access the root then there is a bigger problem
        # and this could result in infinite loop
        if parent == directory:
            break

        directory = parent
    return directory


def _should_include_macos_mount_point(partition: "psutil._common.sdiskpart") -> bool:
    """
    Determine if a macOS mount point should be included in the drive list.

    Args:
        partition: A partition object from psutil.disk_partitions()

    Returns:
        bool: True if the mount point should be included, False otherwise.
    """
    # Skip virtual/system filesystem types:
    # - autofs: Automounter filesystem for automatic mounting/unmounting
    # - devfs: Device filesystem providing access to device files
    # - devtmpfs: Device temporary filesystem (like devfs but in tmpfs)
    # - tmpfs: Temporary filesystem stored in memory
    if partition.fstype in ("autofs", "devfs", "devtmpfs", "tmpfs"):
        return False

    # Skip system volumes under /System/Volumes/ (VM, Preboot, Update, Data, etc.)
    if partition.mountpoint.startswith("/System/Volumes/"):
        return False

    # Include everything else unless it's a system path (/System/, /dev, /private)
    return not partition.mountpoint.startswith(("/System/", "/dev", "/private"))


def _should_include_linux_mount_point(partition: "psutil._common.sdiskpart") -> bool:
    """
    Determine if a Linux/WSL mount point should be included in the drive list.

    Args:
        partition: A partition object from psutil.disk_partitions()

    Returns:
        bool: True if the mount point should be included, False otherwise.
    """
    # Skip virtual/system filesystem types:
    # - autofs: Automounter filesystem for automatic mounting/unmounting
    # - devfs: Device filesystem providing access to device files
    # - devtmpfs: Device temporary filesystem (like devfs but in tmpfs)
    # - tmpfs: Temporary filesystem stored in memory
    # - proc: Process information filesystem
    # - sysfs: System information filesystem
    # - cgroup2: Control group filesystem for resource management
    # - debugfs, tracefs, fusectl, configfs: Kernel debugging/configuration filesystems
    # - securityfs, pstore, bpf: Security and kernel subsystem filesystems
    # - hugetlbfs, mqueue: Specialized system filesystems
    # - devpts: Pseudo-terminal filesystem
    # - binfmt_misc: Binary format support filesystem
    if partition.fstype in (
        "autofs",
        "devfs",
        "devtmpfs",
        "tmpfs",
        "proc",
        "sysfs",
        "cgroup2",
        "debugfs",
        "tracefs",
        "fusectl",
        "configfs",
        "securityfs",
        "pstore",
        "bpf",
        "hugetlbfs",
        "mqueue",
        "devpts",
        "binfmt_misc",
    ):
        return False

    # Skip system paths that users typically don't access:
    # - /dev, /proc, /sys: System directories
    # - /run: Runtime data directory
    # - /boot: Boot partition (typically not accessed by users)
    # - /mnt/wslg: WSL GUI support directory
    # - /mnt/wsl: WSL system integration directory
    # Include everything else (root filesystem, /home, /media, Windows drives in WSL like /mnt/c, etc.)
    return not partition.mountpoint.startswith((
        "/dev",
        "/proc",
        "/sys",
        "/run",
        "/boot",
        "/mnt/wslg",
        "/mnt/wsl",
    ))


def get_mounted_drives() -> list:
    """
    Get a list of mounted drives on the system.

    Returns:
        list: List of mounted drives.
    """
    drives = []
    try:
        # get all partitions
        partitions = psutil.disk_partitions(all=False)

        if os_type == "Windows":
            # For Windows, return the drive letters
            drives = [
                normalise(p.mountpoint)
                for p in partitions
                if p.device and ":" in p.device
            ]
        elif os_type == "Darwin":
            # For macOS, filter out system volumes and keep only user-relevant drives
            drives = [
                p.mountpoint for p in partitions if _should_include_macos_mount_point(p)
            ]
        else:
            # For other Unix-like systems (Linux, WSL, etc.), filter out system mount points
            drives = [
                p.mountpoint for p in partitions if _should_include_linux_mount_point(p)
            ]
    except Exception as e:
        print(f"Error getting mounted drives: {e}")
        print("Using fallback method")
        drives = [path.expanduser("~")]
    return drives



from asyncio import sleep

from pathvalidate import sanitize_filepath
from textual import events, work
from textual.app import ComposeResult
from textual.containers import HorizontalGroup
from textual.content import Content
from textual.screen import ModalScreen
from textual.validation import Length
from textual.widgets import Input, Label

from rovr.functions import icons as icon_utils


class ModalInput(ModalScreen):
    def __init__(
        self,
        border_title: str,
        border_subtitle: str = "",
        initial_value: str = "",
        validators: list = [],
        is_path: bool = False,
        is_folder: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.border_title = border_title
        self.border_subtitle = border_subtitle
        self.initial_value = initial_value
        self.validators = [
            Length(minimum=1, failure_description="A value is required.")
        ] + validators
        self.validators[0].strict = True
        self.is_path = is_path
        self.is_folder = is_folder
        if self.is_path:
            self.icon_widget = Label(
                f"> {icon_utils.get_icon('file', 'default')[0]} ",
                id="icon",
                shrink=True,
            )
        else:
            self.icon_widget = Label("> ", id="icon", shrink=True)

    def compose(self) -> ComposeResult:
        with HorizontalGroup():
            yield self.icon_widget
            yield Input(
                id="input",
                compact=True,
                value=self.initial_value,
                valid_empty=False,
                validators=self.validators,
                validate_on=["changed", "submitted"],
            )

    @work(exclusive=True)
    async def on_input_changed(self, event: Input.Changed) -> None:
        if self.is_path:
            if (
                event.value == self.initial_value and event.value != ""
            ) or self.query_one(Input).is_valid:
                self.icon_widget.classes = "valid"
                self.horizontal_group.classes = "valid"
                self.horizontal_group.border_subtitle = self.border_subtitle
            else:
                self.icon_widget.classes = "invalid"
                self.horizontal_group.classes = "invalid"
                if event.validation_result:
                    self.horizontal_group.border_subtitle = str(
                        event.validation_result.failure_descriptions[0]
                    )
                else:
                    # valid_empty = False
                    self.horizontal_group.border_subtitle = (
                        "The value must not be empty!"
                    )
            if event.value.replace("\\", "/").endswith("/"):
                # dir
                icon = icon_utils.get_icon_for_folder(event.value[:-1])
            elif self.is_folder:
                # dir
                icon = icon_utils.get_icon_for_folder(event.value)
            else:
                # file
                icon = icon_utils.get_icon_for_file(event.value)
            self.icon_widget.update(
                Content.from_markup(f"> [{icon[1]}]{icon[0]}[{icon[1]}] ")
            )

    def on_mount(self) -> None:
        self.horizontal_group: HorizontalGroup = self.query_one(HorizontalGroup)
        inp: Input = self.query_one(Input)
        self.horizontal_group.border_title = self.border_title
        if self.border_subtitle != "":
            self.horizontal_group.border_subtitle = self.border_subtitle
        inp.focus()
        inp.validate(inp.value)
        self.on_input_changed(inp.Changed(inp, inp.value))

    @work
    async def on_input_submitted(self, event: Input.Submitted) -> None:
        """Handle input submission."""
        if (
            not self.query_one(Input).is_valid
            and event.validation_result
            and event.validation_result.failures
        ):
            # shake
            for _ in range(2):
                self.horizontal_group.styles.offset = (1, 0)
                await sleep(0.1)
                self.horizontal_group.styles.offset = (0, 0)
                await sleep(0.1)
            return
        return_path = (
            sanitize_filepath(event.input.value) if self.is_path else event.input.value
        )
        if event.input.value.endswith(("/", "\\")) and not return_path.endswith((
            "/",
            "\\",
        )):
            return_path += "/"
        self.dismiss(return_path)

    def on_key(self, event: events.Key) -> None:
        """Handle escape key to dismiss the dialog."""
        if event.key == "escape":
            event.stop()
            self.dismiss("")



from dataclasses import dataclass, field

from textual.theme import Theme


@dataclass
class RovrThemeClass(Theme):
    name: str
    primary: str
    secondary: str | None = None
    warning: str | None = None
    error: str | None = None
    success: str | None = None
    accent: str | None = None
    foreground: str | None = None
    background: str | None = None
    surface: str | None = None
    panel: str | None = None
    boost: str | None = None
    dark: bool = True
    luminosity_spread: float = 0.15
    text_alpha: float = 0.95
    variables: dict[str, str] = field(default_factory=dict)
    bar_gradient: list[str] | None = None



from os import DirEntry

from textual.content import Content, ContentText
from textual.widgets.option_list import Option
from textual.widgets.selection_list import Selection, SelectionType


class PinnedSidebarOption(Option):
    def __init__(self, icon: list, label: str, *args, **kwargs) -> None:
        super().__init__(
            prompt=Content.from_markup(
                f" [{icon[1]}]{icon[0]}[/{icon[1]}] $name", name=label
            ),
            *args,
            **kwargs,
        )
        self.label = label


class FileListSelectionWidget(Selection):
    # Cache for pre-parsed icon Content objects to avoid repeated markup parsing
    _icon_content_cache: dict[tuple[str, str], Content] = {}

    def __init__(
        self,
        icon: list[str],
        label: str,
        dir_entry: DirEntry,
        value: SelectionType,
        disabled: bool = False,
    ) -> None:
        """
        Initialise the selection.

        Args:
            icon (list[str]): The icon list from a utils function.
            label (str): The label for the option.
            dir_entry (DirEntry): The nt.DirEntry class
            value (SelectionType): The value for the selection.
            disabled (bool) = False: The initial enabled/disabled state. Enabled by default.
        """
        cache_key = (icon[0], icon[1])
        if cache_key not in FileListSelectionWidget._icon_content_cache:
            # Parse the icon markup once and cache it as Content
            FileListSelectionWidget._icon_content_cache[cache_key] = (
                Content.from_markup(f" [{icon[1]}]{icon[0]}[/{icon[1]}] ")
            )

        # Create prompt by combining cached icon content with label
        prompt = FileListSelectionWidget._icon_content_cache[cache_key] + Content(label)

        super().__init__(prompt=prompt, value=value, id=str(value), disabled=disabled)
        self.dir_entry = dir_entry
        self.label = label


class ClipboardSelection(Selection):
    def __init__(self, prompt: ContentText, *args, **kwargs) -> None:
        """
        Initialise the selection.

        Args:
            prompt: The prompt for the selection.
            value: The value for the selection.
            initial_state: The initial selected state of the selection.
            id: The optional ID for the selection.
            disabled: The initial enabled/disabled state. Enabled by default.
        """
        super().__init__(prompt, *args, **kwargs)
        self.initial_prompt = prompt


class KeybindOption(Option):
    def __init__(
        self,
        keys: str,
        description: str,
        max_key_width: int,
        primary_key: str,
        **kwargs,
    ) -> None:
        # Should be named 'label' for searching
        self.label = f" {keys:>{max_key_width}} │ {description} "
        self.key_press = primary_key

        super().__init__(self.label, **kwargs)
        if primary_key == "":
            self.disabled = True


class FinderOption(Option):
    # icon cache
    _icon_content_cache: dict[tuple[str, str], Content] = {}

    def __init__(
        self,
        icon: list[str],
        label: str,
        id: str = "",
        disabled: bool = False,
    ) -> None:
        """
        Initialise the option

        Args:
            icon (list[str]): The icon list from a utils function.
            label (str): The label for the option.
            id (str): The optional id for the option.
            disabled (bool) = False: The initial enabled/disabled state.
        """
        cache_key = (icon[0], icon[1])
        if cache_key not in FinderOption._icon_content_cache:
            # Parse
            FinderOption._icon_content_cache[cache_key] = Content.from_markup(
                f" [{icon[1]}]{icon[0]}[/] "
            )

        # create prompt
        prompt = FinderOption._icon_content_cache[cache_key] + Content(label)
        super().__init__(prompt=prompt, disabled=disabled, id=id)
        self.label = label



from textual import events, work
from textual.app import ComposeResult
from textual.containers import Center, HorizontalGroup
from textual.screen import ModalScreen
from textual.widgets import Label, Static

from rovr.variables.constants import MaxPossible, ascii_logo

max_possible = MaxPossible()


class TerminalTooSmall(ModalScreen):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        yield Static(id="fillerup")
        with Center():
            yield Label(ascii_logo, id="logo")
        with Center():
            with HorizontalGroup(id="height"):
                yield Label("Height: ")
                yield Label(f"[$success]{max_possible.height}[/] > ")
                yield Label("", id="heightThing")
            with HorizontalGroup(id="width"):
                yield Label("Width : ")
                yield Label(f"[$success]{max_possible.width}[/] > ")
                yield Label("", id="widthThing")
        yield Static(id="fillerdown")

    def on_mount(self) -> None:
        for child in self.query("*"):
            child.show_horizontal_scrollbar = False
            child.show_vertical_scrollbar = False
        self.extra_changes()

    @work
    async def extra_changes(self) -> None:
        self.query_one("#heightThing", Label).update(
            f"[${'error' if self.size.height < max_possible.height else 'success'}]{self.size.height}[/]"
        )
        self.query_one("#widthThing", Label).update(
            f"[${'error' if self.size.width < max_possible.width else 'success'}]{self.size.width}[/]"
        )
        for widget in ["#width", "#height", "#fillerup", "#fillerdown", "#logo"]:
            self.query_one(widget).classes = "" if self.size.height > 6 else "hidden"

    @work(exclusive=True)
    async def on_resize(self, event: events.Resize) -> None:
        if (
            event.size.height >= max_possible.height
            and event.size.width >= max_possible.width
        ):
            self.dismiss()
            return
        self.extra_changes()

    def on_key(self, event: events.Key) -> None:
        event.stop()



from os import getcwd, path

from textual import work
from textual.widgets import Button

from rovr.classes import IsValidFilePath
from rovr.functions.icons import get_icon
from rovr.functions.path import normalise
from rovr.screens import ModalInput
from rovr.variables.constants import config


class UnzipButton(Button):
    ALLOW_MAXIMIZE = False

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            get_icon("general", "open")[0],
            classes="option",
            id="unzip",
            *args,
            **kwargs,
        )

    def on_mount(self) -> None:
        if config["interface"]["tooltips"]:
            self.tooltip = "Extract selected archive"

    @work
    async def on_button_pressed(self, event: Button.Pressed) -> None:
        if self.disabled:
            return
        selected_files = await self.app.query_one("#file_list").get_selected_objects()
        if not selected_files or len(selected_files) != 1:
            self.notify(
                "Please select exactly one archive to extract.",
                title="Unzip File",
                severity="warning",
            )
            return

        archive_path = selected_files[0]
        archive_name = path.basename(archive_path)

        default_folder_name = archive_name.rsplit(".", 1)[0]

        response: str = await self.app.push_screen(
            ModalInput(
                border_title="Extract Archive",
                border_subtitle=f"Extract '{archive_name}' to a new folder:",
                initial_value=default_folder_name,
                validators=[IsValidFilePath()],
                is_path=True,
                is_folder=True,
            ),
            wait_for_dismiss=True,
        )

        if not response:
            return

        destination_path = normalise(path.join(getcwd(), response))

        self.app.query_one("ProcessContainer").unzip_file(
            archive_path, destination_path
        )
        self.app.query_one("#file_list").focus()



import stat
import time
from contextlib import suppress
from datetime import datetime
from os import DirEntry, lstat, path, walk

from textual import events, on, work
from textual.containers import VerticalGroup, VerticalScroll
from textual.css.query import NoMatches
from textual.widgets import Static
from textual.worker import WorkerState

from rovr.functions import utils
from rovr.functions.path import is_hidden_file
from rovr.variables.constants import config
from rovr.variables.maps import SPINNER


class MetadataContainer(VerticalScroll):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.current_path: str | None = None
        self._size_worker = None
        self._update_task = None
        self._queued_task = None
        self._queued_task_args: None | DirEntry = None

    def info_of_dir_entry(self, dir_entry: DirEntry, type_string: str) -> str:
        """Get the permission line from a given DirEntry object
        Args:
            dir_entry (DirEntry): The nt.DirEntry class
            type_string (str): The type of file. It should already be handled.
        Returns:
            str: A permission string.
        """
        try:
            file_stat = lstat(dir_entry.path)
        except (OSError, FileNotFoundError):
            return "?????????"
        mode = file_stat.st_mode

        permission_string = ""
        match type_string:
            case "Symlink":
                permission_string = "l"
            case "Directory":
                permission_string = "d"
            case "Junction":
                permission_string = "j"
            case "File":
                permission_string = "-"
            case "Unknown":
                return "???????"

        permission_string += "r" if mode & stat.S_IRUSR else "-"
        permission_string += "w" if mode & stat.S_IWUSR else "-"
        permission_string += "x" if mode & stat.S_IXUSR else "-"

        permission_string += "r" if mode & stat.S_IRGRP else "-"
        permission_string += "w" if mode & stat.S_IWGRP else "-"
        permission_string += "x" if mode & stat.S_IXGRP else "-"

        permission_string += "r" if mode & stat.S_IROTH else "-"
        permission_string += "w" if mode & stat.S_IWOTH else "-"
        permission_string += "x" if mode & stat.S_IXOTH else "-"
        return permission_string

    def any_in_queue(self) -> bool:
        if self._queued_task is not None:
            self._queued_task(self._queued_task_args)
            self._queued_task, self._queued_task_args = None, None
            return True
        return False

    def update_metadata(self, dir_entry: DirEntry) -> None:
        """
        Debounce the update, because some people can be speed travellers
        Args:
            dir_entry (DirEntry): The nt.DirEntry object
        """
        if any(
            worker.is_running
            and worker.node is self
            and worker.name == "_perform_update"
            for worker in self.app.workers
        ):
            self._queued_task = self._perform_update
            self._queued_task_args = dir_entry
        else:
            self._perform_update(dir_entry)

    @work(thread=True)
    def _perform_update(self, dir_entry: DirEntry) -> None:
        """
        After debouncing the update
        Args:
            dir_entry (DirEntry): The nt.DirEntry object
        """
        if self.any_in_queue():
            return
        if not path.exists(dir_entry.path):
            self.app.call_from_thread(self.remove_children)
            self.app.call_from_thread(
                self.mount, Static("Item not found or inaccessible.")
            )
            return

        type_str = "Unknown"
        if dir_entry.is_junction():
            type_str = "Junction"
        elif dir_entry.is_symlink():
            type_str = "Symlink"
        elif dir_entry.is_dir():
            type_str = "Directory"
        elif dir_entry.is_file():
            type_str = "File"
        file_info = self.info_of_dir_entry(dir_entry, type_str)
        # got the type, now we follow
        file_stat = dir_entry.stat()
        is_hidden = is_hidden_file(dir_entry.path)

        values_list = []
        for field in config["metadata"]["fields"]:
            match field:
                case "type":
                    values_list.append(Static(type_str))
                case "permissions":
                    values_list.append(Static(file_info))
                case "hidden":
                    values_list.append(Static("Yes" if is_hidden else "No"))
                case "size":
                    values_list.append(
                        Static(
                            utils.natural_size(
                                file_stat.st_size,
                                config["metadata"]["filesize_suffix"],
                                config["metadata"]["filesize_decimals"],
                            )
                            if type_str == "File"
                            else "--",
                            id="metadata-size",
                        )
                    )
                case "modified":
                    values_list.append(
                        Static(
                            datetime.fromtimestamp(file_stat.st_mtime).strftime(
                                config["metadata"]["datetime_format"]
                            )
                        )
                    )
                case "accessed":
                    values_list.append(
                        Static(
                            datetime.fromtimestamp(file_stat.st_atime).strftime(
                                config["metadata"]["datetime_format"]
                            )
                        )
                    )
                case "created":
                    values_list.append(
                        Static(
                            datetime.fromtimestamp(file_stat.st_ctime).strftime(
                                config["metadata"]["datetime_format"]
                            )
                        )
                    )
        values = VerticalGroup(*values_list, id="metadata-values")

        try:
            for index, child_widget in enumerate(
                self.query_one("#metadata-values").children
            ):
                self.app.call_from_thread(
                    child_widget.update, values_list[index].content
                )
        except NoMatches:
            self.app.call_from_thread(self.remove_children)
            keys_list = []
            for field in config["metadata"]["fields"]:
                match field:
                    case "type":
                        keys_list.append(Static("Type"))
                    case "permissions":
                        keys_list.append(Static("Permissions"))
                    case "hidden":
                        keys_list.append(Static("Hidden"))
                    case "size":
                        keys_list.append(Static("Size"))
                    case "modified":
                        keys_list.append(Static("Modified"))
                    case "accessed":
                        keys_list.append(Static("Accessed"))
                    case "created":
                        keys_list.append(Static("Created"))
            keys = VerticalGroup(*keys_list, id="metadata-keys")
            self.app.call_from_thread(self.mount, keys, values)
        self.current_path = dir_entry.path
        if type_str == "Directory" and self.has_focus:
            self._size_worker = self.calculate_folder_size(dir_entry.path)
        if self.any_in_queue():
            return
        else:
            self._queued_task = None

    @work(thread=True)
    async def calculate_folder_size(self, folder_path: str) -> None:
        """Calculate the size of a folder and update the metadata."""
        size_widget = self.query_one("#metadata-size", Static)
        self.app.call_from_thread(size_widget.update, "Calculating...")

        total_size = 0
        spinner_index = -1
        last_update_time = time.monotonic()
        try:
            for dirpath, _, filenames in walk(folder_path):
                for f in filenames:
                    if self._size_worker is None or self._size_worker.is_cancelled:
                        return
                    fp = path.join(dirpath, f)
                    if not path.islink(fp):
                        with suppress(OSError, FileNotFoundError):
                            total_size += lstat(fp).st_size
                if time.monotonic() - last_update_time > 0.25:
                    spinner_index = spinner_index + 1 if spinner_index // 3 == 0 else 0
                    self.app.call_from_thread(
                        size_widget.update,
                        f"{utils.natural_size(total_size, config['metadata']['filesize_suffix'], config['metadata']['filesize_decimals'])} {SPINNER[spinner_index]}",
                    )
                    last_update_time = time.monotonic()
        except (OSError, FileNotFoundError):
            self.app.call_from_thread(size_widget.update, "Error")
            return

        if not self._size_worker.is_cancelled:
            self.app.call_from_thread(
                size_widget.update,
                utils.natural_size(
                    total_size,
                    config["metadata"]["filesize_suffix"],
                    config["metadata"]["filesize_decimals"],
                ),
            )

    @on(events.Focus)
    def on_focus(self) -> None:
        if self.current_path and path.isdir(self.current_path):
            if self._size_worker:
                return
            self._size_worker = self.calculate_folder_size(self.current_path)

    @on(events.Blur)
    def on_blur(self) -> None:
        if self._size_worker is None or self.app.app_blurred:
            return
        elif self._size_worker.state == WorkerState.SUCCESS:
            self._size_worker = None
        else:
            self._size_worker.cancel()
            self._size_worker = None
            self.set_timer(
                0.1, lambda: self.query_one("#metadata-size", Static).update("--")
            )



from functools import lru_cache
from os import path

from rovr.variables.constants import config
from rovr.variables.maps import (
    ASCII_ICONS,
    ASCII_TOGGLE_BUTTON_ICONS,
    FILE_MAP,
    FILES_MAP,
    FOLDER_MAP,
    ICONS,
    TOGGLE_BUTTON_ICONS,
)


@lru_cache(maxsize=128)
def get_icon_for_file(location: str) -> list:
    """
    Get the icon and color for a file based on its name or extension.

    Args:
        location (str): The name or path of the file.

    Returns:
        list: The icon and color for the file.
    """
    if not config["interface"]["nerd_font"]:
        return ASCII_ICONS["file"]["default"]
    file_name = path.basename(location).lower()

    # 0. Check for custom icons if configured
    if "icons" in config and "files" in config["icons"]:
        for custom_icon in config["icons"]["files"]:
            pattern = custom_icon["pattern"].lower()
            match_type = custom_icon.get("match_type", "exact")

            is_match = False
            if (
                match_type == "exact"
                and file_name == pattern
                or match_type == "endswith"
                and file_name.endswith(pattern)
            ):
                is_match = True

            if is_match:
                return [custom_icon["icon"], custom_icon["color"]]

    # 1. Check for full filename match
    if file_name in FILES_MAP:
        icon_key = FILES_MAP[file_name]
        return ICONS["file"].get(icon_key, ICONS["file"]["default"])

    # 2. Check for extension match
    if "." in file_name:
        # This is for hidden files like `.gitignore`
        extension = "." + file_name.split(".")[-1]
        if extension in FILE_MAP:
            icon_key = FILE_MAP[extension]
            return ICONS["file"].get(icon_key, ICONS["file"]["default"])

    # 3. Default icon
    return ICONS["file"]["default"]


@lru_cache(maxsize=128)
def get_icon_for_folder(location: str) -> list:
    """Get the icon and color for a folder based on its name.

    Args:
        location (str): The name or path of the folder.

    Returns:
        list: The icon and color for the folder.
    """
    folder_name = path.basename(location).lower()

    if not config["interface"]["nerd_font"]:
        return ASCII_ICONS["folder"].get(folder_name, ASCII_ICONS["folder"]["default"])

    # 0. Check for custom icons if configured
    if "icons" in config and "folders" in config["icons"]:
        for custom_icon in config["icons"]["folders"]:
            pattern = custom_icon["pattern"].lower()
            match_type = custom_icon.get("match_type", "exact")

            is_match = False
            if (
                match_type == "exact"
                and folder_name == pattern
                or match_type == "endswith"
                and folder_name.endswith(pattern)
            ):
                is_match = True

            if is_match:
                return [custom_icon["icon"], custom_icon["color"]]

    # Check for special folder types
    if folder_name in FOLDER_MAP:
        icon_key = FOLDER_MAP[folder_name]
        return ICONS["folder"].get(icon_key, ICONS["folder"]["default"])
    else:
        return ICONS["folder"]["default"]


@lru_cache(maxsize=128)
def get_icon(outer_key: str, inner_key: str) -> list:
    """Get an icon from double keys.
    Args:
        outer_key (str): The category name (general/folder/file)
        inner_key (str): The icon's name

    Returns:
        list[str,str]: The icon and color for the icon
    """
    if not config["interface"]["nerd_font"]:
        return ASCII_ICONS.get(outer_key, {"empty": None}).get(inner_key, " ")
    else:
        return ICONS[outer_key][inner_key]


@lru_cache(maxsize=128)
def get_toggle_button_icon(key: str) -> str:
    if not config["interface"]["nerd_font"]:
        return ASCII_TOGGLE_BUTTON_ICONS[key]
    else:
        return TOGGLE_BUTTON_ICONS[key]



import bz2
import gzip
import lzma
import tarfile
import zipfile
from pathlib import Path
from types import TracebackType
from typing import IO, List, Literal, Optional, Union

import rarfile


class Archive:
    """Unified handler for ZIP, TAR and RAR files with context manager support."""

    def __init__(
        self,
        filename: Union[str, Path],
        mode: str = "r",
        compression_level: Optional[int] = None,
    ) -> None:
        """Initialize the archive handler.

        Args:
            filename: Path to the archive file
            mode: File access mode ('r' for read, 'w' for write, 'a' for append)
            compression_level: Compression level (ZIP: 0-9, TAR gzip: 0-9, TAR bzip2: 1-9)
                             If None, uses default compression

        Raises:
            ValueError: If mode is not supported or compression_level is out of range
        """  # noqa: DOC502
        self.filename = str(filename)
        self.mode = mode
        self.compression_level = compression_level
        self._archive: Optional[
            Union[zipfile.ZipFile, tarfile.TarFile, rarfile.RarFile]
        ] = None
        self._is_zip: Optional[bool] = None
        self._is_rar: Optional[bool] = None

    def __enter__(self) -> "Archive":
        """Context manager entry - opens the archive.

        Returns:
            Self for method chaining in with statement

        Raises:
            FileNotFoundError: If the archive file doesn't exist (for read mode)
            zipfile.BadZipFile: If ZIP file is corrupted
            tarfile.TarError: If TAR file is corrupted or unreadable
            rarfile.BadRarFile: If RAR file is corrupted
        """  # noqa: DOC502
        self._detect_and_open()
        return self

    def __exit__(
        self,
        exc_type: Optional[type],
        exc_val: Optional[Exception],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Context manager exit - closes the archive.

        Args:
            exc_type: Exception type if an exception occurred
            exc_val: Exception value if an exception occurred
            exc_tb: Traceback if an exception occurred
        """
        """Context manager exit - closes the archive."""
        if self._archive:
            self._archive.close()

    def _detect_and_open(self) -> None:
        """Detect file type and open appropriate handler.

        Raises:
            FileNotFoundError: If the archive file doesn't exist (for read mode)
            zipfile.BadZipFile: If ZIP file is corrupted
            tarfile.TarError: If TAR file is corrupted or format not supported
            rarfile.BadRarFile: If RAR file is corrupted
            ValueError: If file extension is not recognized or compression_level is invalid
        """  # noqa: DOC502
        filename_lower = self.filename.lower()

        if filename_lower.endswith(".zip"):
            self._is_zip = True
            self._is_rar = False
            if self.compression_level is not None:
                if not (0 <= self.compression_level <= 9):
                    raise ValueError("ZIP compression level must be between 0-9")
                self._archive = zipfile.ZipFile(
                    self.filename, self.mode, compresslevel=self.compression_level
                )
            else:
                self._archive = zipfile.ZipFile(self.filename, self.mode)
            # Check for password protection
            if self.mode == "r":
                assert isinstance(self._archive, zipfile.ZipFile)
                if any(zinfo.flag_bits & 0x1 for zinfo in self._archive.infolist()):
                    self._archive.close()
                    raise ValueError("Password-protected ZIP files are not supported")
        elif filename_lower.endswith(".rar"):
            self._is_zip = False
            self._is_rar = True
            if self.mode != "r":
                raise ValueError("RAR files can only be opened in read mode ('r')")
            self._archive = rarfile.RarFile(self.filename, self.mode)
            # Check for password protection
            assert isinstance(self._archive, rarfile.RarFile)
            if self._archive.needs_password():
                self._archive.close()
                raise ValueError("Password-protected RAR files are not supported")
        else:
            # Assume it's a tar file - let tarfile auto-detect compression
            self._is_zip = False
            self._is_rar = False
            if self.mode == "r":
                self._archive = tarfile.open(self.filename, "r:*")  # noqa: SIM115
            else:
                tar_mode = self._get_tar_write_mode()
                if self.compression_level is not None:
                    self._archive = self._open_tar_with_compression(tar_mode)
                else:
                    self._archive = tarfile.open(self.filename, tar_mode)  # noqa: SIM115

    def _get_tar_write_mode(self) -> Literal["w:gz", "w:bz2", "w:xz", "w"]:
        """Determine tar write mode based on file extension.

        Returns:
            Appropriate tarfile mode string for writing
        """
        filename_lower = self.filename.lower()
        if filename_lower.endswith((".tar.gz", ".tgz")):
            return "w:gz"
        elif filename_lower.endswith((".tar.bz2", ".tbz2")):
            return "w:bz2"
        elif filename_lower.endswith(".tar.xz"):
            return "w:xz"
        else:
            return "w"

    def _open_tar_with_compression(
        self, tar_mode: Literal["w:gz", "w:bz2", "w:xz", "w"]
    ) -> tarfile.TarFile:
        """Open TAR file with specified compression level.

        Args:
            tar_mode: TAR mode string (e.g., 'w:gz', 'w:bz2')

        Returns:
            Opened TarFile with compression level applied

        Raises:
            ValueError: If compression level is invalid for the compression type
        """
        if ":gz" in tar_mode:
            # Gzip compression: level 0-9
            if not (0 <= self.compression_level <= 9):
                raise ValueError("Gzip compression level must be between 0-9")
            gz_file = gzip.open(  # noqa: SIM115
                self.filename, self.mode + "b", compresslevel=self.compression_level
            )
            return tarfile.open(fileobj=gz_file, mode="w")

        elif ":bz2" in tar_mode:
            # Bzip2 compression: level 1-9
            if not (1 <= self.compression_level <= 9):
                raise ValueError("Bzip2 compression level must be between 1-9")
            bz2_file = bz2.open(  # noqa: SIM115
                self.filename, self.mode + "b", compresslevel=self.compression_level
            )
            return tarfile.open(fileobj=bz2_file, mode="w")

        elif ":xz" in tar_mode:
            # XZ/LZMA compression: level 0-9 (but typically 1-9 is useful)
            if not (0 <= self.compression_level <= 9):
                raise ValueError("XZ compression level must be between 0-9")
            xz_file = lzma.open(  # noqa: SIM115
                self.filename, self.mode + "b", preset=self.compression_level
            )
            return tarfile.open(fileobj=xz_file, mode="w")

        else:
            # Uncompressed tar - compression level ignored
            return tarfile.open(self.filename, tar_mode)

    def infolist(
        self,
    ) -> List[Union[zipfile.ZipInfo, tarfile.TarInfo, rarfile.RarInfo]]:
        """Return list of archive members (similar to zipfile.infolist()).

        Returns:
            List of ZipInfo, TarInfo or RarInfo objects

        Raises:
            RuntimeError: If archive is not opened
        """
        if not self._archive:
            raise RuntimeError("Archive not opened")

        if self._is_zip or self._is_rar:
            return self._archive.infolist()
        else:
            return self._archive.getmembers()

    def namelist(self) -> List[str]:
        """Return list of member names.

        Returns:
            List of strings containing all member file/directory names in the archive

        Raises:
            RuntimeError: If archive is not opened
        """
        if not self._archive:
            raise RuntimeError("Archive not opened")

        if self._is_zip or self._is_rar:
            return self._archive.namelist()
        else:
            return self._archive.getnames()

    def extract(
        self,
        member: Union[str, zipfile.ZipInfo, tarfile.TarInfo, rarfile.RarInfo],
        path: Optional[Union[str, Path]] = None,
    ) -> str:
        """Extract a single member to the specified path.

        Args:
            member: Name of the file to extract, or ZipInfo/TarInfo/RarInfo object
            path: Directory to extract to. If None, extracts to current directory

        Returns:
            Path to the extracted file

        Raises:
            RuntimeError: If archive is not opened
            KeyError: If member is not found in the archive
            OSError: If extraction fails due to filesystem issues
        """  # noqa: DOC502
        if not self._archive:
            raise RuntimeError("Archive not opened")

        if self._is_rar:
            # rarfile's extract method returns None. We construct the path manually.
            member_filename = (
                member.filename if isinstance(member, rarfile.RarInfo) else member
            )
            self._archive.extract(member, path)
            return str(Path(path or ".") / member_filename)

        return self._archive.extract(member, path)

    def open(
        self,
        member: Union[str, zipfile.ZipInfo, tarfile.TarInfo, rarfile.RarInfo],
        mode: str = "r",
    ) -> Optional[IO[bytes]]:
        """Open a member file for reading.

        Args:
            member: Name of the file to open, or ZipInfo/TarInfo/RarInfo object
            mode: File open mode (only 'r' supported for TAR and RAR files)

        Returns:
            File-like object for reading the member's contents, or None if member
            is a directory or cannot be opened

        Raises:
            RuntimeError: If archive is not opened
            KeyError: If member is not found in the archive
            ValueError: If mode is not supported for TAR/RAR files
        """  # noqa: DOC502
        if not self._archive:
            raise RuntimeError("Archive not opened")

        if self._is_zip:
            return self._archive.open(member, mode)
        elif self._is_rar:
            if mode != "r":
                raise ValueError("RAR members can only be opened in read mode ('r')")
            return self._archive.open(member, mode)
        else:
            # For tar files, use extractfile
            return self._archive.extractfile(member)



from typing import ClassVar

from rich.segment import Segment
from rich.style import Style
from textual import events, work
from textual.binding import BindingType
from textual.content import Content
from textual.strip import Strip
from textual.widgets import Button, SelectionList
from textual.widgets.option_list import OptionDoesNotExist

from rovr.classes import ClipboardSelection
from rovr.functions import icons as icon_utils
from rovr.functions import path as path_utils
from rovr.variables.constants import config, vindings


class Clipboard(SelectionList, inherit_bindings=False):
    """A selection list that displays the clipboard contents."""

    BINDINGS: ClassVar[list[BindingType]] = list(vindings)

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.clipboard_contents = []

    def on_mount(self) -> None:
        self.paste_button: Button = self.app.query_one("#paste")
        self.paste_button.disabled = True
        self.file_list = self.app.query_one("#file_list")

    async def copy_to_clipboard(self, items: list[str]) -> None:
        """Copy the selected files to the clipboard"""
        self.deselect_all()
        for item in items[::-1]:
            self.insert_selection_at_beginning(
                ClipboardSelection(
                    prompt=Content(
                        f"{icon_utils.get_icon('general', 'copy')[0]} {item}"
                    ),
                    value=path_utils.compress(f"{item}-copy"),
                    initial_state=True,
                )
            )
        self.refresh(layout=True)
        self.paste_button.disabled = False

    async def cut_to_clipboard(self, items: list[str]) -> None:
        self.deselect_all()
        """Cut the selected files to the clipboard."""
        for item in items[::-1]:
            if isinstance(item, str):
                self.insert_selection_at_beginning(
                    ClipboardSelection(
                        prompt=Content(
                            f"{icon_utils.get_icon('general', 'cut')[0]} {item}"
                        ),
                        value=path_utils.compress(f"{item}-cut"),
                        initial_state=True,
                    )
                )
        self.refresh(layout=True)
        self.paste_button.disabled = False
        for item_number in range(len(items)):
            self.select(self.get_option_at_index(item_number))

    # Use better versions of the checkbox icons
    def _get_left_gutter_width(
        self,
    ) -> int:
        """Returns the size of any left gutter that should be taken into account.

        Returns:
            The width of the left gutter.
        """
        return len(
            icon_utils.get_toggle_button_icon("left")
            + icon_utils.get_toggle_button_icon("inner")
            + icon_utils.get_toggle_button_icon("right")
            + " "
        )

    def render_line(self, y: int) -> Strip:
        """Render a line in the display.

        Args:
            y: The line to render.

        Returns:
            A [`Strip`][textual.strip.Strip] that is the line to render.
        """
        line = super(SelectionList, self).render_line(y)

        _, scroll_y = self.scroll_offset
        selection_index = scroll_y + y
        try:
            selection = self.get_option_at_index(selection_index)
        except OptionDoesNotExist:
            return line

        component_style = "selection-list--button"
        if selection.value in self._selected:
            component_style += "-selected"
        if self.highlighted == selection_index:
            component_style += "-highlighted"

        underlying_style = next(iter(line)).style or self.rich_style
        assert underlying_style is not None

        button_style = self.get_component_rich_style(component_style)

        side_style = Style.from_color(button_style.bgcolor, underlying_style.bgcolor)

        side_style += Style(meta={"option": selection_index})
        button_style += Style(meta={"option": selection_index})

        return Strip([
            Segment(icon_utils.get_toggle_button_icon("left"), style=side_style),
            Segment(
                icon_utils.get_toggle_button_icon("inner_filled")
                if selection.value in self._selected
                else icon_utils.get_toggle_button_icon("inner"),
                style=button_style,
            ),
            Segment(icon_utils.get_toggle_button_icon("right"), style=side_style),
            Segment(" ", style=underlying_style),
            *line,
        ])

    # Why isnt this already a thing
    def insert_selection_at_beginning(self, content: ClipboardSelection) -> None:
        """Insert a new selection at the beginning of the clipboard list.

        Args:
            content (ClipboardSelection): A pre-created Selection object to insert.
        """
        # Check for duplicate ID
        if content.id is not None and content.id in self._id_to_option:
            self.remove_option(content.id)

        # insert
        self._options.insert(0, content)

        # update self._values
        values = {content.value: 0}

        # update mapping
        for option, index in list(self._option_to_index.items()):
            self._option_to_index[option] = index + 1
        for key, value in self._values.items():
            values[key] = value + 1
        self._values = values
        self._option_to_index[content] = 0

        # update id mapping
        if content.id is not None:
            self._id_to_option[content.id] = content

        # force redraw
        self._clear_caches()

        # since you insert at beginning, highlighted should go down
        if self.highlighted is not None:
            self.highlighted += 1

        # redraw
        # self.refresh(layout=True)

    @work
    async def on_key(self, event: events.Key) -> None:
        if self.has_focus:
            if event.key in config["keybinds"]["delete"]:
                """Delete the selected files from the clipboard."""
                if self.highlighted is None:
                    self.notify(
                        "No files selected to delete from the clipboard.",
                        title="Clipboard",
                        severity="warning",
                    )
                    return
                self.remove_option_at_index(self.highlighted)
                self.paste_button.disabled = self.add_options
                if self.option_count == 0:
                    return
                event.stop()
            elif event.key in config["keybinds"]["toggle_all"]:
                """Select all items in the clipboard."""
                if len(self.selected) == len(self.options):
                    self.deselect_all()
                else:
                    self.select_all()
                event.stop()



import contextlib
from subprocess import TimeoutExpired, run

from textual import events, work
from textual.app import ComposeResult
from textual.containers import VerticalGroup
from textual.content import Content
from textual.screen import ModalScreen
from textual.widgets import Input, OptionList
from textual.widgets.option_list import Option

from rovr.functions import path as path_utils
from rovr.variables.constants import config


class ZoxideOptionList(OptionList):
    async def _on_click(self, event: events.Click) -> None:
        """React to the mouse being clicked on an item.

        Args:
            event: The click event.
        """
        event.prevent_default()
        clicked_option: int | None = event.style.meta.get("option")
        if clicked_option is not None and not self._options[clicked_option].disabled:
            if event.chain == 2:
                if self.highlighted != clicked_option:
                    self.highlighted = clicked_option
                self.action_select()
            else:
                self.highlighted = clicked_option


class ZDToDirectory(ModalScreen):
    """Screen with a dialog to z to a directory, using zoxide"""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._queued_task = None
        self._queued_task_args: str | None = None

    def compose(self) -> ComposeResult:
        with VerticalGroup(id="zoxide_group", classes="zoxide_group"):
            yield Input(
                id="zoxide_input",
                placeholder="Enter directory name or pattern",
            )
            yield ZoxideOptionList(
                Option("  No input provided", disabled=True),
                id="zoxide_options",
                classes="empty",
            )

    def on_mount(self) -> None:
        self.zoxide_input: Input = self.query_one("#zoxide_input")
        self.zoxide_input.border_title = "zoxide"
        self.zoxide_input.focus()
        self.zoxide_options: ZoxideOptionList = self.query_one("#zoxide_options")
        self.zoxide_options.border_title = "Folders"
        self.zoxide_options.can_focus = False
        self.zoxide_updater(Input.Changed(self.zoxide_input, value=""))

    def on_input_changed(self, event: Input.Changed) -> None:
        if any(
            worker.is_running and worker.node is self for worker in self.app.workers
        ):
            self._queued_task = self.zoxide_updater
            self._queued_task_args = event
        else:
            self.zoxide_updater(event=event)

    def any_in_queue(self) -> bool:
        if self._queued_task is not None:
            self._queued_task(self._queued_task_args)
            self._queued_task, self._queued_task_args = None, None
            return True
        return False

    def _parse_zoxide_line(
        self, line: str, show_scores: bool
    ) -> tuple[str, str | None]:
        line = line.strip()
        if not show_scores:
            return line, None

        # Example "  <floating_score> <path_with_spaces>"
        # Split only on first space to make sure path with spaces work
        parts = line.split(None, 1)
        if len(parts) == 2:
            score_str, path = parts
            return path, score_str
        else:
            # This should ideally never happen
            self.notify(
                # Not printing the entire line as that could be too big for UI
                # message. We anyway have the lines in logs
                "Unexpected tokens count while parsing zoxide lines",
                title="Zoxide Plugin",
                severity="error",
            )
            print(f"Problems while parsing zoxide line - '{line}'")
            return line, None

    @work(thread=True)
    def zoxide_updater(self, event: Input.Changed) -> None:
        """Update the list"""
        search_term = event.value.strip()
        # check 1 for queue, to ignore subprocess as a whole
        if self.any_in_queue():
            return

        zoxide_cmd = ["zoxide", "query", "--list"]
        show_scores = config["plugins"]["zoxide"].get("show_scores", False)
        if show_scores:
            zoxide_cmd.append("--score")
        zoxide_cmd.append("--")

        zoxide_cmd += search_term.split()

        try:
            zoxide_output = run(zoxide_cmd, capture_output=True, text=True, timeout=3)
        except (OSError, TimeoutExpired) as exc:
            # zoxide not installed
            if self.any_in_queue():
                return
            self.app.call_from_thread(self.zoxide_options.clear_options)
            self.app.call_from_thread(
                self.zoxide_options.add_option,
                Option(
                    "  zoxide is missing on $PATH or cannot be executed"
                    if isinstance(exc, OSError)
                    else "  zoxide took too long to respond",
                    disabled=True,
                ),
            )
            self.any_in_queue()
            return
        # check 2 for queue, to ignore mounting as a whole
        if self.any_in_queue():
            return
        zoxide_options: ZoxideOptionList = self.query_one(
            "#zoxide_options", ZoxideOptionList
        )
        options = []
        if zoxide_output.stdout:
            first_score_width = 0
            for line in zoxide_output.stdout.splitlines():
                path, score = self._parse_zoxide_line(line, show_scores)
                if show_scores and score:
                    # This ensures that we only add necessary padding
                    # first score is going to be the largest, so we take its width
                    if first_score_width == 0:
                        first_score_width = len(score)
                    # Fixed size to make it look good.
                    display_text = f" {score:>{first_score_width}} │ {path}"
                else:
                    display_text = f" {path}"

                # Use original path for ID (not display text)
                options.append(
                    Option(Content(display_text), id=path_utils.compress(path))
                )
            if len(options) == len(zoxide_options.options) and all(
                options[i].id == zoxide_options.options[i].id
                for i in range(len(options))
            ):  # ie same~ish query, resulting in same result
                pass
            else:
                # unline normally, I'm using an add_option**s** function
                # using it without has a likelihood of DuplicateID being
                # raised, or just nothing showing up. By having the clear
                # options and add options functions nearby, it hopefully
                # reduces the likelihood of an empty option list
                self.app.call_from_thread(zoxide_options.clear_options)
                self.app.call_from_thread(zoxide_options.add_options, options)
                self.app.call_from_thread(zoxide_options.remove_class, "empty")
                zoxide_options.highlighted = 0
        else:
            # No Matches to the query text
            self.app.call_from_thread(zoxide_options.clear_options)
            self.app.call_from_thread(
                zoxide_options.add_option,
                Option("  --No matches found--", disabled=True),
            )
            self.app.call_from_thread(zoxide_options.add_class, "empty")
        # check 3, if somehow there's a new request after the mount
        if self.any_in_queue():
            return  # nothing much to do now

    def on_input_submitted(self, event: Input.Submitted) -> None:
        if any(
            worker.is_running and worker.node is self for worker in self.app.workers
        ):
            return
        if self.zoxide_options.highlighted is None:
            self.zoxide_options.highlighted = 0
        if self.zoxide_options.highlighted_option.disabled:
            return
        self.zoxide_options.action_select()

    # You can't manually tab into the option list, but you can click, so I guess
    @work(exclusive=True)
    async def on_option_list_option_selected(
        self, event: ZoxideOptionList.OptionSelected
    ) -> None:
        """Handle option selection."""
        selected_value = event.option.id
        assert selected_value is not None
        # ignore if zoxide got uninstalled, why are you doing this
        with contextlib.suppress(TimeoutExpired, OSError):
            run(
                ["zoxide", "add", path_utils.decompress(selected_value)],
                capture_output=True,
                text=True,
                timeout=1,
            )
        if selected_value and not event.option.disabled:
            self.dismiss(selected_value)
        else:
            self.dismiss(None)

    def on_key(self, event: events.Key) -> None:
        """Handle key presses."""
        match event.key:
            case "escape":
                event.stop()
                self.dismiss(None)
            case "down":
                event.stop()
                zoxide_options = self.query_one("#zoxide_options")
                if zoxide_options.options:
                    zoxide_options.action_cursor_down()
            case "up":
                event.stop()
                zoxide_options = self.query_one("#zoxide_options")
                if zoxide_options.options:
                    zoxide_options.action_cursor_up()
            case "tab":
                event.stop()
                self.focus_next()
            case "shift+tab":
                event.stop()
                self.focus_previous()

    def on_option_list_option_highlighted(
        self, event: ZoxideOptionList.OptionHighlighted
    ) -> None:
        if (
            self.zoxide_options.option_count == 0
            or self.zoxide_options.get_option_at_index(0).disabled
        ):
            self.zoxide_options.border_subtitle = "0/0"
        else:
            self.zoxide_options.border_subtitle = f"{self.zoxide_options.highlighted + 1}/{self.zoxide_options.option_count}"



import os
from os import path

import jsonschema
import toml
import ujson
from lzstring import LZString
from rich.console import Console

from rovr.functions.utils import deep_merge
from rovr.variables.maps import (
    VAR_TO_DIR,
)

lzstring = LZString()
pprint = Console().print


def load_config() -> tuple[dict, dict]:
    """
    Load both the template config and the user config

    Returns:
        dict: the config
    """

    if not path.exists(VAR_TO_DIR["CONFIG"]):
        os.makedirs(VAR_TO_DIR["CONFIG"])
    if not path.exists(path.join(VAR_TO_DIR["CONFIG"], "config.toml")):
        with open(path.join(VAR_TO_DIR["CONFIG"], "config.toml"), "w") as file:
            file.write(
                '#:schema  https://raw.githubusercontent.com/NSPC911/rovr/refs/heads/master/src/rovr/config/schema.json\n[theme]\ndefault = "nord"'
            )

    with open(path.join(path.dirname(__file__), "../config/config.toml"), "r") as f:
        try:
            template_config = toml.loads(f.read())
        except toml.decoder.TomlDecodeError as e:
            pprint(f"[bright_red]TOML Syntax Error:\n    {e}")
            exit(1)

    user_config_path = path.join(VAR_TO_DIR["CONFIG"], "config.toml")
    user_config = {}
    if path.exists(user_config_path):
        with open(user_config_path, "r") as f:
            user_config_content = f.read()
            if user_config_content:
                user_config = toml.loads(user_config_content)
    # Don't really have to consider the else part, because it's created further down
    config = deep_merge(template_config, user_config)
    # check with schema
    with open(path.join(path.dirname(__file__), "../config/schema.json"), "r") as f:
        schema = ujson.load(f)

    # fix schema with 'required' keys
    def add_required_recursively(node: dict) -> None:
        if isinstance(node, dict):
            if (
                node.get("type") == "object" and "properties" in node
            ) and "required" not in node:
                node["required"] = list(node["properties"].keys())
            for key in node:
                add_required_recursively(node[key])
        elif isinstance(node, list):
            for item in node:
                add_required_recursively(item)

    add_required_recursively(schema)

    try:
        jsonschema.validate(config, schema)
    except jsonschema.exceptions.ValidationError as exception:
        # pprint(exception.__dict__)
        path_str = "root"
        if exception.path:
            path_str = ".".join(str(p) for p in exception.path)
        pprint(
            f"[underline bright_red]Config Error[/] at path [bold cyan]{path_str}[/]:"
        )
        match exception.validator:
            case "required":
                pprint(f"{exception.message}, but is not provided.")
            case "type":
                type_error_message = (
                    f"Invalid type: expected [yellow]{exception.validator_value}[/yellow], "
                    f"but got [yellow]{type(exception.instance).__name__}[/yellow]."
                )
                pprint(type_error_message)
            case "enum":
                enum_error_message = (
                    f"Invalid value [yellow]'{exception.instance}'[/yellow]. "
                    f"\nAllowed values are: {exception.validator_value}"
                )
                pprint(enum_error_message)
            case _:
                pprint(f"[yellow]{exception.message}[/yellow]")
        exit(1)

    # slight config fixes
    # image protocol because "AutoImage" doesn't work with Sixel
    if config["settings"]["image_protocol"] == "Auto":
        config["settings"]["image_protocol"] = ""
    return schema, config


def config_setup() -> None:
    # check config folder
    if not path.exists(VAR_TO_DIR["CONFIG"]):
        os.makedirs(VAR_TO_DIR["CONFIG"])
    # Textual doesn't seem to have a way to check whether the
    # CSS file exists while it is in operation, but textual
    # only craps itself when it can't find it as the app starts
    # so no issues
    if not path.exists(path.join(VAR_TO_DIR["CONFIG"], "style.tcss")):
        with open(path.join(VAR_TO_DIR["CONFIG"], "style.tcss"), "a") as _:
            pass



from humanize import naturalsize
from lzstring import LZString
from rich.console import Console
from textual.widget import Widget

from rovr.variables.maps import (
    BORDER_BOTTOM,
)

lzstring = LZString()
pprint = Console().print


def deep_merge(d: dict, u: dict) -> dict:
    """Mini lodash merge
    Args:
        d (dict): old dictionary
        u (dict): new dictionary, to merge on top of d

    Returns:
        dict: Merged dictionary
    """
    for k, v in u.items():
        if isinstance(v, dict):
            d[k] = deep_merge(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def set_nested_value(d: dict, path_str: str, value: bool) -> None:
    """Sets a value in a nested dictionary using a dot-separated path string.

    Args:
        d (dict): The dictionary to modify.
        path_str (str): The dot-separated path to the key (e.g., "plugins.bat").
        value (bool): The value to set. (boolean for now)
    """
    keys = path_str.split(".")
    current = d
    for i, key in enumerate(keys):
        if i == len(keys) - 1:
            try:
                if isinstance(current[key], dict) and "enabled" in current[key]:
                    current[key]["enabled"] = value
                elif type(current[key]) is type(value):
                    current[key] = value
                else:
                    pprint("[bright_red underline]Config Error:[/]")
                    pprint(
                        f"[cyan bold]{path_str}[/]'s new value of type [cyan b]{type(value).__name__}[/] is not a [bold cyan]{type(current[key]).__name__}[/] type, and cannot be modified."
                    )
                    exit(1)
            except KeyError:
                pprint("[bright_red underline]Config Error:[/]")
                pprint(
                    f"[cyan b]{path_str}[/] is not a valid path to an existing value and hence cannot be set."
                )
                exit(1)
        else:
            if not isinstance(current.get(key), dict):
                current[key] = {}
            current = current[key]


def set_scuffed_subtitle(element: Widget, *sections: str) -> None:
    """The most scuffed way to display a custom subtitle

    Args:
        element (Widget): The element containing style information.
        *sections (str): The sections to display
    """
    border_bottom = BORDER_BOTTOM.get(
        element.styles.border_bottom[0], BORDER_BOTTOM["blank"]
    )
    subtitle = ""
    for index, section in enumerate(sections):
        subtitle += section
        if index + 1 != len(sections):
            subtitle += " "
            subtitle += (
                border_bottom if element.app.ansi_color else f"[r]{border_bottom}[/]"
            )
            subtitle += " "

    element.border_subtitle = subtitle


def natural_size(integer: int, suffix: str, filesize_decimals: int) -> str:
    assert suffix in ["decimal", "binary", "gnu"]
    match suffix:
        case "gnu":
            return naturalsize(
                value=integer,
                gnu=True,
                format=f"%.{filesize_decimals}f",
            )
        case "binary":
            return naturalsize(
                value=integer,
                binary=True,
                format=f"%.{filesize_decimals}f",
            )
        case _:
            return naturalsize(value=integer, format=f"%.{filesize_decimals}f")



from os import getcwd, path

from textual.widgets import Button

from rovr.functions.icons import get_icon


class BackButton(Button):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(get_icon("general", "left")[0], id="back", *args, **kwargs)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Go back in the sesison's history"""
        if self.disabled:
            return
        state = self.app.tabWidget.active_tab.session
        if state.historyIndex != 0:
            state.historyIndex -= 1
        # ! reminder to add a check for path!
        self.app.cd(
            state.directories[state.historyIndex]["path"],
            add_to_history=False,
        )


class ForwardButton(Button):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(get_icon("general", "right")[0], id="forward", *args, **kwargs)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Go forward in the session's history"""
        if self.disabled:
            return
        state = self.app.tabWidget.active_tab.session
        state.historyIndex += 1
        # ! reminder to add a check for path!
        self.app.cd(
            state.directories[state.historyIndex]["path"],
            add_to_history=False,
        )


class UpButton(Button):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(get_icon("general", "up")[0], id="up", *args, **kwargs)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Go up the current location's directory"""
        if self.disabled:
            return
        parent = getcwd().split(path.sep)[-1]
        self.app.cd(
            path.sep.join(getcwd().split(path.sep)[:-1]) + path.sep, focus_on=parent
        )



import contextlib

from textual import events, work
from textual.css.query import NoMatches
from textual.fuzzy import Matcher
from textual.types import OptionDoesNotExist
from textual.widgets import Input, OptionList, SelectionList
from textual.widgets.option_list import Option
from textual.widgets.selection_list import Selection, SelectionError

from rovr.functions.utils import set_scuffed_subtitle


class SearchInput(Input):
    def __init__(self, always_add_disabled: bool = True, *args, **kwargs) -> None:
        super().__init__(
            *args, password=False, compact=True, select_on_focus=False, **kwargs
        )
        self.always_add_disabled = always_add_disabled
        self.selected = set()

    def on_mount(self) -> None:
        self.items_list = self.parent.query_one(OptionList)
        if isinstance(self.items_list, SelectionList):
            self.item_list_type = "Selection"
        elif isinstance(self.items_list, OptionList):
            self.item_list_type = "Option"
        else:
            raise NoMatches(
                f"type {type(self.items_list).__name__} was matched but expected either OptionList or SelectionList"
            )

    # exclusive when too many options, and not enough time to mount
    @work(exclusive=True)
    async def on_input_changed(self, event: Input.Changed) -> None:
        if self.item_list_type == "Selection":
            assert isinstance(self.items_list, SelectionList)
            self.selected.update({*self.items_list.selected})
        else:
            assert isinstance(self.items_list, OptionList)
        try:
            highlighted = self.items_list.highlighted_option.id
        except AttributeError:
            highlighted = None
        self.app.tabWidget.active_tab.session.search = event.value
        if event.value == "":
            self.items_list.clear_options()
            self.items_list.add_options(self.items_list.list_of_options)
            if highlighted is not None:
                with contextlib.suppress(OptionDoesNotExist, SelectionError):
                    self.items_list.highlighted = self.items_list.get_option_index(
                        highlighted
                    )
            else:
                self.items_list.highlighted = 0
            if self.item_list_type == "Selection":
                for option_id in self.selected:
                    with contextlib.suppress(OptionDoesNotExist):
                        if not self.items_list.select_mode_enabled:
                            with self.items_list.prevent(
                                self.items_list.SelectedChanged
                            ):
                                self.items_list.select(
                                    self.items_list.get_option(option_id)
                                )
                        else:
                            self.items_list.select(
                                self.items_list.get_option(option_id)
                            )
            return
        self.items_list.clear_options()
        matches = []
        matcher = Matcher(
            event.value,
        )
        assert hasattr(self.items_list, "list_of_options")
        for option in self.items_list.list_of_options:
            assert isinstance(option, Option)
            if self.always_add_disabled and option.disabled:
                matches.append(option)
                continue
            score = matcher.match(option.label)
            if score > 0:
                matches.append(option)
        if matches:
            self.items_list.add_options(matches)
        else:
            if self.item_list_type == "Option":
                self.items_list.add_option(
                    Option("   --no-matches--", id="", disabled=True)
                )
            else:
                self.items_list.add_option(
                    Selection("   --no-matches--", value="", id="", disabled=True)
                )
                set_scuffed_subtitle(
                    self.items_list.parent,
                    "SELECT" if self.items_list.select_mode_enabled else "NORMAL",
                    "0/0",
                )
        if self.items_list.highlighted is None:
            if highlighted is not None:
                try:
                    self.items_list.highlighted = self.items_list.get_option_index(
                        highlighted
                    )
                except OptionDoesNotExist:
                    self.items_list.action_cursor_down()
            else:
                self.items_list.action_cursor_down()
        if self.item_list_type == "Selection":
            for option_id in self.selected:
                with contextlib.suppress(OptionDoesNotExist):
                    if not self.items_list.select_mode_enabled:
                        with self.items_list.prevent(self.items_list.SelectedChanged):
                            self.items_list.select(
                                self.items_list.get_option(option_id)
                            )
                    else:
                        self.items_list.select(self.items_list.get_option(option_id))

    def on_input_submitted(self, event: Input.Submitted) -> None:
        self.items_list.focus()

    def on_key(self, event: events.Key) -> None:
        if event.key == "escape":
            self.items_list.focus()
            event.stop()

    def clear_selected(self) -> None:
        self.selected = set()



import platform
from dataclasses import dataclass

from textual.binding import Binding, BindingType

from rovr.functions.config import config_setup, load_config

# Initialize the config once at import time
if "config" not in globals():
    global config, schema
    schema, config = load_config()
    config_setup()


@dataclass
class PreviewContainerTitles:
    image = "Image Preview"
    bat = "File Preview (bat)"
    file = "File Preview"
    folder = "Folder Preview"
    archive = "Archive Preview"


buttons_that_depend_on_path = [
    "#copy",
    "#cut",
    "#rename",
    "#delete",
    "#zip",
    "#copy_path",
]

ascii_logo = r"""
 ___ ___ _ _ ___
|  _| . | | |  _|
|_| |___|\_/|_|"""


class MaxPossible:
    @property
    def height(self) -> int:
        return 13 if config["interface"]["use_reactive_layout"] else 24

    @property
    def width(self) -> int:
        return 26 if config["interface"]["use_reactive_layout"] else 70


vindings: list[BindingType] = (
    [
        Binding(bind, "cursor_down", "Down", show=False)
        for bind in config["keybinds"]["down"]
    ]
    + [Binding(bind, "last", "Last", show=False) for bind in config["keybinds"]["end"]]
    + [
        Binding(bind, "select", "Select", show=False)
        for bind in config["keybinds"]["down_tree"]
    ]
    + [
        Binding(bind, "first", "First", show=False)
        for bind in config["keybinds"]["home"]
    ]
    + [
        Binding(bind, "page_down", "Page Down", show=False)
        for bind in config["keybinds"]["page_down"]
    ]
    + [
        Binding(bind, "page_up", "Page Up", show=False)
        for bind in config["keybinds"]["page_up"]
    ]
    + [
        Binding(bind, "cursor_up", "Up", show=False)
        for bind in config["keybinds"]["up"]
    ]
)

os_type: str = platform.system()



from textual import events
from textual.app import ComposeResult
from textual.containers import Grid, HorizontalGroup, VerticalGroup
from textual.screen import ModalScreen
from textual.widgets import Button, Label, Switch


class YesOrNo(ModalScreen):
    """Screen with a dialog that asks whether you accept or deny"""

    def __init__(
        self,
        message: str,
        reverse_color: bool = False,
        with_toggle: bool = False,
        border_title: str = "",
        border_subtitle: str = "",
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.message = message
        self.reverse_color = reverse_color
        self.with_toggle = with_toggle
        self.border_title = border_title
        self.border_subtitle = border_subtitle

    def compose(self) -> ComposeResult:
        with Grid(id="dialog"):
            with VerticalGroup(id="question_container"):
                for message in self.message.splitlines():
                    yield Label(message, classes="question")
            yield Button(
                "\\[Y]es",
                variant="error" if self.reverse_color else "primary",
                id="yes",
            )
            yield Button(
                "\\[N]o", variant="primary" if self.reverse_color else "error", id="no"
            )
            if self.with_toggle:
                with HorizontalGroup(id="dontAskAgain"):
                    yield Switch()
                    yield Label("Don't \\[a]sk again")

    def on_mount(self) -> None:
        self.query_one("#dialog").classes = "with_toggle" if self.with_toggle else ""
        self.query_one("#dialog").border_title = self.border_title
        self.query_one("#dialog").border_subtitle = self.border_subtitle

    def on_key(self, event: events.Key) -> None:
        """Handle key presses."""
        match event.key.lower():
            case "y":
                event.stop()
                self.dismiss(
                    {"value": True, "toggle": self.query_one(Switch).value}
                    if self.with_toggle
                    else True
                )
            case "n" | "escape":
                event.stop()
                self.dismiss(
                    {"value": False, "toggle": self.query_one(Switch).value}
                    if self.with_toggle
                    else False
                )
            case "a":
                event.stop()
                self.query_one(Switch).action_toggle_switch()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.dismiss(
            {"value": event.button.id == "yes", "toggle": self.query_one(Switch).value}
            if self.with_toggle
            else event.button.id == "yes"
        )



import os
from os import path

import ujson

from rovr.variables.maps import (
    VAR_TO_DIR,
)

from .path import normalise

pins = {}


def load_pins() -> dict:
    """
    Load the pinned files from a JSON file in the user's config directory.
    Returns:
        dict: A dictionary with the default values, and the custom added pins.
    """
    global pins
    user_pins_file_path = path.join(VAR_TO_DIR["CONFIG"], "pins.json")

    # Ensure the user's config directory exists
    if not path.exists(VAR_TO_DIR["CONFIG"]):
        os.makedirs(VAR_TO_DIR["CONFIG"])
    if not path.exists(user_pins_file_path):
        pins = {
            "default": [
                {"name": "Home", "path": "$HOME"},
                {"name": "Downloads", "path": "$DOWNLOADS"},
                {"name": "Documents", "path": "$DOCUMENTS"},
                {"name": "Desktop", "path": "$DESKTOP"},
                {"name": "Pictures", "path": "$PICTURES"},
                {"name": "Videos", "path": "$VIDEOS"},
                {"name": "Music", "path": "$MUSIC"},
            ],
            "pins": [],
        }
        try:
            with open(user_pins_file_path, "w") as f:
                ujson.dump(pins, f, escape_forward_slashes=False, indent=2)
        except IOError:
            pass

    try:
        with open(user_pins_file_path, "r") as f:
            pins = ujson.load(f)
    except (IOError, ValueError):
        # Reset pins on corrupt or something else happened
        pins = {
            "default": [
                {"name": "Home", "path": "$HOME"},
                {"name": "Downloads", "path": "$DOWNLOADS"},
                {"name": "Documents", "path": "$DOCUMENTS"},
                {"name": "Desktop", "path": "$DESKTOP"},
                {"name": "Pictures", "path": "$PICTURES"},
                {"name": "Videos", "path": "$VIDEOS"},
                {"name": "Music", "path": "$MUSIC"},
            ],
            "pins": [],
        }

    # If list died
    if "default" not in pins or not isinstance(pins["default"], list):
        pins["default"] = [
            {"name": "Home", "path": "$HOME"},
            {"name": "Downloads", "path": "$DOWNLOADS"},
            {"name": "Documents", "path": "$DOCUMENTS"},
            {"name": "Desktop", "path": "$DESKTOP"},
            {"name": "Pictures", "path": "$PICTURES"},
            {"name": "Videos", "path": "$VIDEOS"},
            {"name": "Music", "path": "$MUSIC"},
        ]
    if "pins" not in pins or not isinstance(pins["pins"], list):
        pins["pins"] = []

    for section_key in ["default", "pins"]:
        for item in pins[section_key]:
            if (
                isinstance(item, dict)
                and "path" in item
                and isinstance(item["path"], str)
            ):
                # Expand variables
                for var, dir_path_val in VAR_TO_DIR.items():
                    item["path"] = item["path"].replace(f"${var}", dir_path_val)
                # Normalize to forward slashes
                item["path"] = normalise(item["path"])
    return pins


def add_pin(pin_name: str, pin_path: str | bytes) -> None:
    """
    Add a pin to the pins file.

    Args:
        pin_name (str): Name of the pin.
        pin_path (str): Path of the pin.
    """
    global pins

    pins_to_write = ujson.loads(ujson.dumps(pins))

    pin_path_normalized = normalise(pin_path)
    pins_to_write.setdefault("pins", []).append({
        "name": pin_name,
        "path": pin_path_normalized,
    })

    sorted_vars = sorted(VAR_TO_DIR.items(), key=lambda x: len(x[1]), reverse=True)
    for section_key in ["default", "pins"]:
        if section_key in pins_to_write:
            for item in pins_to_write[section_key]:
                if (
                    isinstance(item, dict)
                    and "path" in item
                    and isinstance(item["path"], str)
                ):
                    for var, dir_path_val in sorted_vars:
                        item["path"] = item["path"].replace(dir_path_val, f"${var}")

    try:
        user_pins_file_path = path.join(VAR_TO_DIR["CONFIG"], "pins.json")
        with open(user_pins_file_path, "w") as f:
            ujson.dump(pins_to_write, f, escape_forward_slashes=False, indent=2)
    except IOError:
        pass

    load_pins()


def remove_pin(pin_path: str | bytes) -> None:
    """
    Remove a pin from the pins file.

    Args:
        pin_path (str): Path of the pin to remove.
    """
    global pins

    pins_to_write = ujson.loads(ujson.dumps(pins))

    pin_path_normalized = normalise(pin_path)
    if "pins" in pins_to_write:
        pins_to_write["pins"] = [
            pin
            for pin in pins_to_write["pins"]
            if not (isinstance(pin, dict) and pin.get("path") == pin_path_normalized)
        ]

    sorted_vars = sorted(VAR_TO_DIR.items(), key=lambda x: len(x[1]), reverse=True)
    for section_key in ["default", "pins"]:
        if section_key in pins_to_write:
            for item in pins_to_write[section_key]:
                if (
                    isinstance(item, dict)
                    and "path" in item
                    and isinstance(item["path"], str)
                ):
                    for var, dir_path_val in sorted_vars:
                        item["path"] = item["path"].replace(dir_path_val, f"${var}")

    try:
        user_pins_file_path = path.join(VAR_TO_DIR["CONFIG"], "pins.json")
        with open(user_pins_file_path, "w") as f:
            ujson.dump(pins_to_write, f, escape_forward_slashes=False, indent=2)
    except IOError:
        pass

    load_pins()  # Reload


def toggle_pin(pin_name: str, pin_path: str) -> None:
    """
    Toggle a pin in the pins file. If it exists, remove it; if not, add it.

    Args:
        pin_name (str): Name of the pin.
        pin_path (str): Path of the pin.
    """
    pin_path_normalized = normalise(pin_path)

    pin_exists = False
    if "pins" in pins:
        for pin_item in pins["pins"]:
            if (
                isinstance(pin_item, dict)
                and pin_item.get("path") == pin_path_normalized
            ):
                pin_exists = True
                break

    if pin_exists:
        remove_pin(pin_path_normalized)
    else:
        add_pin(pin_name, pin_path_normalized)



from textual import events
from textual.app import ComposeResult
from textual.containers import Grid, HorizontalGroup, VerticalGroup
from textual.screen import ModalScreen
from textual.widgets import Button, Label, Switch


class GiveMePermission(ModalScreen):
    def __init__(
        self, message: str, border_title: str = "", border_subtitle: str = "", **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.message = message
        self.border_title = border_title
        self.border_subtitle = border_subtitle

    def compose(self) -> ComposeResult:
        with Grid(id="dialog"):
            with VerticalGroup(id="question_container"):
                for message in self.message.splitlines():
                    yield Label(message, classes="question")
            yield Button("\\[Y]es", id="force", variant="error")
            yield Button("\\[N]o/\\[S]kip", id="skip", variant="warning")
            with HorizontalGroup():
                yield Button("\\[C]ancel", id="cancel", variant="default")
            with HorizontalGroup(id="dontAskAgain"):
                yield Switch()
                yield Label("Don't \\[a]sk again")

    def on_mount(self) -> None:
        self.query_one("#dialog").border_title = self.border_title
        self.query_one("#dialog").border_subtitle = self.border_subtitle

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.dismiss({"value": event.button.id, "toggle": self.query_one(Switch).value})

    def on_key(self, event: events.Key) -> None:
        match event.key.lower():
            case "y":
                event.stop()
                self.dismiss({"value": "force", "toggle": self.query_one(Switch).value})
            case "n" | "s":
                event.stop()
                self.dismiss({"value": "skip", "toggle": self.query_one(Switch).value})
            case "escape" | "c":
                event.stop()
                self.dismiss({
                    "value": "cancel",
                    "toggle": self.query_one(Switch).value,
                })
            case "a":
                event.stop()
                self.query_one(Switch).action_toggle_switch()



try:
    import rich_click as click
    from rich import box
    from rich.table import Table

    from rovr.functions.path import normalise
    from rovr.functions.utils import pprint, set_nested_value
    from rovr.variables.constants import config
    from rovr.variables.maps import VAR_TO_DIR

    click.rich_click.USE_RICH_MARKUP = True
    click.rich_click.USE_MARKDOWN = False
    click.rich_click.SHOW_ARGUMENTS = True
    click.rich_click.GROUP_ARGUMENTS_OPTIONS = False
    click.rich_click.MAX_WIDTH = 88
    click.rich_click.STYLE_OPTION = "bold cyan"
    click.rich_click.STYLE_ARGUMENT = "bold cyan"
    click.rich_click.STYLE_COMMAND = "bold cyan"
    click.rich_click.STYLE_SWITCH = "bold cyan"
    click.rich_click.STYLE_METAVAR = "bold yellow"
    click.rich_click.STYLE_METAVAR_SEPARATOR = "dim"
    click.rich_click.STYLE_USAGE = "bold cyan"
    click.rich_click.STYLE_USAGE_COMMAND = "bold"
    click.rich_click.STYLE_HELPTEXT_FIRST_LINE = ""
    click.rich_click.STYLE_HELPTEXT = "dim"
    click.rich_click.STYLE_OPTION_DEFAULT = "dim magenta"
    click.rich_click.STYLE_REQUIRED_SHORT = "red"
    click.rich_click.STYLE_REQUIRED_LONG = "dim red"
    click.rich_click.STYLE_OPTIONS_PANEL_BORDER = "blue bold"
    click.rich_click.STYLE_COMMANDS_PANEL_BORDER = "white"

    @click.command(help="A post-modern terminal file explorer")
    @click.option(
        "--with",
        "with_features",
        multiple=True,
        type=str,
        help="Enable a feature (e.g., 'plugins.bat').",
    )
    @click.option(
        "--without",
        "without_features",
        multiple=True,
        type=str,
        help="Disable a feature (e.g., 'interface.tooltips').",
    )
    @click.option(
        "--config-path",
        "show_config_path",
        multiple=False,
        type=bool,
        default=False,
        is_flag=True,
        help="Show the path to the config folder.",
    )
    @click.option(
        "--version",
        "show_version",
        multiple=False,
        type=bool,
        default=False,
        is_flag=True,
        help="Show the current version of rovr.",
    )
    @click.option(
        "--cwd-file",
        "cwd_file",
        multiple=False,
        type=str,
        default="",
        help="Write the final working directory to this file on exit.",
    )
    @click.option(
        "--chooser-file",
        "chooser_file",
        multiple=False,
        type=str,
        default="",
        help="Write chosen file(s) (newline-separated) to this file on exit.",
    )
    @click.option_panel("Config", options=["--with", "--without"])
    @click.option_panel("Paths", options=["--chooser-file", "--cwd-file"])
    @click.option_panel(
        "Miscellaneous", options=["--version", "--config-path", "--help"]
    )
    @click.argument("path", type=str, required=False, default="")
    @click.rich_config({"show_arguments": True})
    def main(
        with_features: list[str],
        without_features: list[str],
        show_config_path: bool,
        show_version: bool,
        cwd_file: str,
        chooser_file: str,
        path: str,
    ) -> None:
        """A post-modern terminal file explorer"""

        if show_config_path:
            from pathlib import Path

            table = Table(title="", border_style="blue", box=box.ROUNDED)
            table.add_column("type")
            table.add_column("path")
            path_config = Path(VAR_TO_DIR["CONFIG"])
            if path_config.is_relative_to(Path.home()):
                config_path = "~/" + normalise(
                    str(path_config.relative_to(Path.home()))
                )
            else:
                config_path = path_config
            table.add_row("[cyan]custom config[/]", f"{config_path}/config.toml")
            table.add_row("[yellow]pinned folders[/]", f"{config_path}/pins.json")
            table.add_row("[hot_pink]custom styles[/]", f"{config_path}/style.tcss")
            if config["settings"]["cd_on_quit"]:
                table.add_row(
                    "[green]path saved on quit[/]",
                    f"{config_path}/rovr_cd_on_quit",
                )
            pprint(table)
            return
        elif show_version:
            pprint("v0.3.0")
            return

        for feature_path in with_features:
            set_nested_value(config, feature_path, True)

        for feature_path in without_features:
            set_nested_value(config, feature_path, False)

        from rovr.app import Application

        # TODO: Need to move this 'path' in the config dict, or a new runtime_config dict
        # Eventually there will be many options coming via arguments, but we cant keep sending all of
        # them via this Application's __init__ function here
        Application(
            watch_css=True,
            startup_path=path,
            cwd_file=cwd_file if cwd_file else None,
            chooser_file=chooser_file if chooser_file else None,
        ).run()

except KeyboardInterrupt:
    pass



import asyncio
from os import path
from typing import ClassVar

from textual import events, on, work
from textual.binding import BindingType
from textual.widgets import Input, OptionList
from textual.widgets.option_list import Option

from rovr.classes import FolderNotFileError, PinnedSidebarOption
from rovr.functions import icons as icon_utils
from rovr.functions import path as path_utils
from rovr.functions import pins as pin_utils
from rovr.variables.constants import config, vindings


class PinnedSidebar(OptionList, inherit_bindings=False):
    DRIVE_WATCHER_FREQUENCY: float = config["settings"]["drive_watcher_frequency"]
    # Just so that I can disable space
    BINDINGS: ClassVar[list[BindingType]] = list(vindings)

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    @work
    async def reload_pins(self) -> None:
        """Reload pins shown

        Raises:
            FolderNotFileError: If the pin location is a file, and not a folder.
        """
        # be extra sure
        available_pins = pin_utils.load_pins()
        pins = available_pins["pins"]
        default = available_pins["default"]
        self.list_of_options = []
        print(f"Reloading pins: {available_pins}")
        print(f"Reloading default folders: {default}")
        self.clear_options()
        for default_folder in default:
            if not path.isdir(default_folder["path"]):
                if path.exists(default_folder["path"]):
                    raise FolderNotFileError(
                        f"Expected a folder but got a file: {default_folder['path']}"
                    )
                else:
                    pass
            if "icon" in default_folder:
                icon = default_folder["icon"]
            elif path.isdir(default_folder["path"]):
                icon = icon_utils.get_icon_for_folder(default_folder["name"])
            else:
                icon = icon_utils.get_icon_for_file(default_folder["name"])
            self.list_of_options.append(
                PinnedSidebarOption(
                    icon=icon,
                    label=default_folder["name"],
                    id=f"{path_utils.compress(default_folder['path'])}-default",
                )
            )
        self.list_of_options.append(
            Option(" Pinned", id="pinned-header", disabled=True)
        )
        for pin in pins:
            try:
                pin["path"]
            except KeyError:
                break
            if not path.isdir(pin["path"]):
                if path.exists(pin["path"]):
                    raise FolderNotFileError(
                        f"Expected a folder but got a file: {pin['path']}"
                    )
                else:
                    pass
            if "icon" in pin:
                icon = pin["icon"]
            elif path.isdir(pin["path"]):
                icon = icon_utils.get_icon_for_folder(pin["name"])
            else:
                icon = icon_utils.get_icon_for_file(pin["name"])
            self.list_of_options.append(
                PinnedSidebarOption(
                    icon=icon,
                    label=pin["name"],
                    id=f"{path_utils.compress(pin['path'])}-pinned",
                )
            )
        self.list_of_options.append(
            Option(" Drives", id="drives-header", disabled=True)
        )
        drives = path_utils.get_mounted_drives()
        for drive in drives:
            self.list_of_options.append(
                PinnedSidebarOption(
                    icon=icon_utils.get_icon("folder", ":/drive:"),
                    label=drive,
                    id=f"{path_utils.compress(drive)}-drives",
                )
            )
        self.add_options(self.list_of_options)

    @work
    async def watch_for_drive_changes_and_update(self) -> None:
        self._drives = path_utils.get_mounted_drives()
        while True:
            await asyncio.sleep(self.DRIVE_WATCHER_FREQUENCY)
            try:
                new_drives = path_utils.get_mounted_drives()
                if self._drives != new_drives:
                    self._drives = new_drives
                    self.reload_pins()
            except Exception as e:
                print(
                    f"Exception of type {type(e).__name__} while watching drives: {e}"
                )
                continue

    async def on_mount(self) -> None:
        """Reload the pinned files from the config."""
        self.input: Input = self.parent.query_one(Input)
        self.reload_pins()
        self.watch_for_drive_changes_and_update()

    @on(events.Enter)
    @work
    async def show_input_when_hover(self, event: events.Focus) -> None:
        self.input.add_class("show")

    @on(events.Leave)
    @work
    async def hide_input_when_leave(self, event: events.Leave) -> None:
        self.input.remove_class("show")

    @on(events.Focus)
    def focus_this_thing(self, event: events.Focus) -> None:
        if self.highlighted is None:
            self.action_cursor_down()

    async def on_option_list_option_selected(
        self, event: OptionList.OptionSelected
    ) -> None:
        """Handle the selection of an option in the pinned sidebar.
        Args:
            event (OptionList.OptionSelected): The event

        Raises:
            FolderNotFileError: If the pin found is a file and not a folder.
        """
        selected_option = event.option
        # Get the file path from the option id
        assert selected_option.id is not None
        file_path = path_utils.decompress(selected_option.id.split("-")[0])
        if not path.isdir(file_path):
            if path.exists(file_path):
                raise FolderNotFileError(
                    f"Expected a folder but got a file: {file_path}"
                )
            else:
                return
        self.app.cd(file_path)
        self.app.query_one("#file_list").focus()
        with self.input.prevent(Input.Changed):
            self.input.clear()
        self.clear_options()
        self.add_options(self.list_of_options)

    def on_key(self, event: events.Key) -> None:
        if event.key in config["keybinds"]["focus_search"]:
            self.input.focus()
